# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
import time
import os
import threading
import csv
import numpy as np
from typing import List, Optional
from brainflow.board_shim import BoardShim, BrainFlowInputParams

from octopus_sensing.common.message_creators import MessageType
from octopus_sensing.devices.monitored_device import MonitoredDevice
from octopus_sensing.devices.common import SavingModeEnum


class BrainFlowStreaming(MonitoredDevice):
    '''
    Manage brainflow streaming

    Attributes
    ----------
    device_id
        Device ID. 
        Brainflow support a list of devices, to see supported device IDs go to:
        https://brainflow.readthedocs.io/en/stable/SupportedBoards.html
    
    sampling_rate
        the sampling rate for recording data
    
    brain_flow_input_params
        Each supported board in brainflow gets some parameters, to see the list
        of parameters for each board go to:
        https://brainflow.readthedocs.io/en/stable/SupportedBoards.html
    
    name
        device name
        This name will be used in the output path to identify each device's data
    
    output_path
                 The path for recording files.
                 Audio files will be recorded in folder {output_path}/{name}

    saving_mode
        The way of saving data. It saves data continiously in a file 
        or saves data which are related to various stimulus in separate files. 
        default is SavingModeEnum.CONTINIOUS_SAVING_MODE
        SavingModeEnum is [CONTINIOUS_SAVING_MODE, SEPARATED_SAVING_MODE]
    ** kwargs:
       Extra optional arguments according to the board type
    
    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`
    
    Examples
    ---------
    Here is an example of using brainflow for reading cyton_daisy board data

    >>> params = BrainFlowInputParams()
    >>> params.serial_port = "/dev/ttyUSB0"
    >>> my_brainflow = 
    ...       BrainFlowStreaming(2,
    ...                          125,
    ...                          brain_flow_input_params=params,
    ...                          name="cyton_daisy",
    ...                          output_path="./output",
    ...                          saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE)
        
    '''
    def __init__(self,
                 device_id: int,
                 sampling_rate: int,
                 brain_flow_input_params: BrainFlowInputParams,
                 saving_mode: int=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                 name: Optional[str] = None,
                 output_path: str = "output"):
        super().__init__(name=name, output_path=output_path)

        self._saving_mode = saving_mode
        self._stream_data: List[float] = []
        self.sampling_rate = sampling_rate

        self._board = None
        self._device_id = device_id
        self._brain_flow_input_params = brain_flow_input_params
        self._terminate = False
        self._trigger = None
        self._experiment_id = None

        self.output_path = os.path.join(self.output_path, self.name)
        os.makedirs(self.output_path, exist_ok=True)
        self._state = ""

    def _run(self):
        self._board = BoardShim(self._device_id, self._brain_flow_input_params)
        self._board.set_log_level(0)
        self._board.prepare_session()

        threading.Thread(target=self._stream_loop).start()

        while True:
            message = self.message_queue.get()
            if message is None:
                continue
            if message.type == MessageType.START:
                if self._state == "START":
                    print("Brainflow streaming has already recorded the START triger")
                else:
                    print("Brainflow start")
                    self.__set_trigger(message)
                    self._experiment_id = message.experiment_id
                    self._state = "START"
            elif message.type == MessageType.STOP:
                if self._state == "STOP":
                    print("Brainflow streaming has already recorded the STOP triger")
                else:
                    print("Brainflow stop")
                    if self._saving_mode == SavingModeEnum.SEPARATED_SAVING_MODE:
                        self._experiment_id = message.experiment_id
                        file_name = \
                            "{0}/{1}-{2}-{3}.csv".format(self.output_path,
                                                        self.name,
                                                        self._experiment_id,
                                                        message.stimulus_id)
                        self._save_to_file(file_name)
                        self._stream_data = []
                    else:
                        self._experiment_id = message.experiment_id
                        self.__set_trigger(message)
                    self._state = "STOP"
            elif message.type == MessageType.TERMINATE:
                self._terminate = True
                if self._saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
                    file_name = \
                        "{0}/{1}-{2}.csv".format(self.output_path,
                                                 self.name,
                                                 self._experiment_id)
                    self._save_to_file(file_name)
                break

        self._board.stop_stream()

    def _stream_loop(self):
        self._board.start_stream()
        while True:
            if self._terminate is True:
                break
            data = self._board.get_board_data()
            if np.array(data).shape[1] != 0:
                self._stream_data.extend(list(np.transpose(data)))
                last_record = self._stream_data.pop()
                last_record = list(last_record)
                now = str(datetime.now().time())
                last_record.append(now)
                last_record.append(time.time())
                if self._trigger is not None:
                    last_record.append(self._trigger)
                    self._trigger = None
                self._stream_data.append(last_record)
            else:
                time.sleep(0.1)
            #    print("brainflow: didn't read any data")

    def __set_trigger(self, message):
        '''
        Takes a message and set the trigger using its data
        
        Parameters
        ----------
        message: Message
            a message object
        '''
        self._trigger = \
            "{0}-{1}-{2}".format(message.type,
                                 message.experiment_id,
                                 str(message.stimulus_id).zfill(2))

    def _save_to_file(self, file_name):
        with open(file_name, 'a') as csv_file:
            writer = csv.writer(csv_file)
            for row in self._stream_data:
                writer.writerow(row)
                csv_file.flush()

    def _get_monitoring_data(self):
        '''Returns latest collected data for monitoring/visualizing purposes.'''
        # Last three seconds
        return self._stream_data[-1 * 3 * self.sampling_rate:]



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os
from octopus_sensing.stimuli.stimulus import Stimulus

class VideoStimulus(Stimulus):
    '''
    Creates video stimulus

    Attributes
    -----------


    Parameters
    ----------
    id: str
        The id of stimulus

    path: str
        The path of video
    '''

    def __init__(self, id, path):
        super().__init__(id, path)

    def show(self):
        '''
        Displays video stimulus using VLC
        '''
        os.system("vlc --fullscreen --play-and-exit {0}".format(self.path))


# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.
import gi
gi.require_version('Gtk', '3.0')  # nopep8
from gi.repository import Gtk  # nopep8
from typing import List, Union, Any

from octopus_sensing.questionnaire.question import Question

FONT_STYLE = "<span font_desc='Tahoma 16'>{}</span>"


class Option():
    '''
    The class for creating opinions

    Attributes
    ----------

    Parameters
    ----------

    id: int
        A unique ID for the opinion
    
    label: str
        The text of opinion
    
    value: Any
        The value of opinion

    '''

    def __init__(self, id: int, label: str=None, value: Any=None):
        self.id = id
        self.label = label
        if self.label is None:
            self.label = str(id)
        self.value = value
        if self.value is None:
            self.value = id


class OpinionQuestion(Question):
    '''
    The class for creating opinion questions using Gtk 3.0


    Attributes
    ----------

    Parameters
    ----------
    id: str
        A unique ID for the question

    text: str
        The text of question
    
    options: Union[dict, int]

    image_path: str, default: None
        The path to an image which will be showed with the question.
    

    default_answer: Union[int, str], default: 0
        The default checked option
    '''
    def __init__(self, id: str, text: str, options: Union[dict, int],
                 image_path: str = None, default_answer: int = 0):
        super().__init__(id, text)

        self._options = []
        if isinstance(options, int):
            for i in range(options):
                option = Option(i, label=str(i), value=i)
                self._options.append(option)
        else:
            i = 0
            for key, value in options.items():
                option = Option(i, label=key, value=value)
                self._options.append(option)
                i += 1

        self._image_path = image_path
        self.answer = default_answer
    
    def render(self, grid: Gtk.Grid, grid_row: int):
        '''
        renders a question for adding to a questionnaire

        Parameters
        ----------
        grid: Gtk.Grid
            a Gtk grid object that this question will be added to it
            
        grid_row: int
            The row number of grid that the question will be added to it
        
        Returns
        -------
        row_counter: int
            The grid's row for adding the next object after adding this question
        
        Examples
        --------
        Creating an opinion question and adding it to the questionnaire

        >>> emotions = {"Happiness": 4, "Sadness": 6, "Neutral": 5, "Fear": 3, "Anger": 1}
        >>> question_1 = OpinionQuestion("q1",
        ...                              "1- What emotion did you feel the most?",
        ...                              options=emotions,
        ...                              default_answer=5)
        >>> questionnaire = Questionnaire("after_stimuli",
        ...                               "study01_p10",
        ...                               "stimuli00",
        ...                               "After Stimulus Questionnaire")
        >>> questionnaire.add_questions([question_1])

        See Also
        -----------
        :class:`octopus_sensing.questionnaire.questionnaire`

        '''

        row_counter = grid_row

        # Question box
        question_label_box = Gtk.Box(spacing=120)
        question_label = Gtk.Label()
        question_label.set_markup(FONT_STYLE.format(self._text))
        question_label_box.pack_start(question_label, False, False, 0)

        grid.attach(question_label_box, 0, row_counter, 1, 1)
        row_counter += 1

        # Image box
        image_box = None
        if self._image_path is not None:
            image_box = Gtk.Box(spacing=120)
            image = Gtk.Image.new_from_file(self._image_path)
            image_box.pack_start(image, False, False, 0)
            grid.attach(image_box, 0, row_counter, 1, 1)
            row_counter += 1

        # Options box
        options_box = Gtk.Box(spacing=120)
        option_buttons: List[Gtk.RadioButton] = []
        for i, option in enumerate(self._options):
            if i == 0:
                option_button = \
                    Gtk.RadioButton.new_with_label_from_widget(None, str(option.id))
            else:
                option_button = \
                    Gtk.RadioButton.new_with_label_from_widget(option_buttons[0],
                                                               str(option.id))
            option_button.connect("toggled",
                                  self.__on_option_button_toggled,
                                  option.value)
            option_button.get_child().set_markup(FONT_STYLE.format(option.label))
            option_buttons.append(option_button)
            options_box.pack_start(option_button, False, False, 0)
            if option.value == self.answer:
                option_button.set_active(True)
        grid.attach(options_box, 0, row_counter, 1, 1)
        row_counter += 1
        return row_counter

    def __on_option_button_toggled(self, button, name):
        if button.get_active():
            self.answer = name

    def get_answer(self) -> int:
        '''
        Gets selected answer

        Returns
        -------

        answer: int
            answer
        '''
        return self.answer



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os
import os.path
import queue
import random
import time
import tempfile
import pickle
import struct
import http.client
import multiprocessing
import multiprocessing.dummy
import multiprocessing.queues

import pytest
import pyOpenBCI
import serial


class MockSample:
    def __init__(self, channels):
        self.channels_data = [round(random.uniform(
            0.01, 0.9), 5) for _ in range(channels)]
        self.aux_data = [round(random.uniform(0.01, 0.9), 5)
                         for _ in range(channels)]
        self.id = random.randrange(1, 200)


class MockedOpenBCICyton:
    def __init__(self, daisy=True):
        self._channels = 16
        if daisy is False:
            self._channels = 8

    def stop_stream(self):
        pass

    def start_stream(self, callback):
        # Four seconds of data
        for _ in range(4 * 128):
            callback(MockSample(self._channels))
            # Sample rate: 128 per second
            time.sleep(1 / 128)


class MockedSerial:
    def __init__(self, port, speed):
        pass

    def flushInput(self):
        pass

    def write(self, data):
        pass

    def close(self):
        pass

    def read(self, size):
        if size == 1:
            # Asking for ack
            return struct.pack('B', 0xff)
        elif size == 9:
            # Enquiry configurations
            # Only byte 7 (number of channels) and byte 8 (buffer size) is used by the app.
            return b"\x00\x00\x00\x00\x00\x00\x00\x05\xff"
        elif size == 5:
            # Channels
            return b"\x65\x66\x67\x68\x69"
        elif size == 14:
            # One frame of data
            frame = []
            for _ in range(14):
                # 48 to 57 are numbers 0 to 9
                frame.append(random.randint(48, 57))
            # Sample rate: 128 per second
            time.sleep(1 / 128)
            return bytes(frame)


@pytest.fixture(scope="module")
def mocked():
    # Preventing processes from creating a new process
    original_process = multiprocessing.Process
    multiprocessing.Process = multiprocessing.dummy.Process
    original_queue = multiprocessing.queues.Queue
    multiprocessing.queues.Queue = queue.Queue
    original_queue_method = multiprocessing.Queue
    multiprocessing.Queue = queue.Queue

    original_openbcicyton = pyOpenBCI.OpenBCICyton
    pyOpenBCI.OpenBCICyton = MockedOpenBCICyton

    original_serial = serial.Serial
    serial.Serial = MockedSerial

    yield None

    multiprocessing.Process = original_process
    multiprocessing.queues.Queue = original_queue
    multiprocessing.Queue = original_queue_method
    pyOpenBCI.OpenBCICyton = original_openbcicyton
    serial.Serial = original_serial


def test_system_health(mocked):
    '''Runs the whole system to roughly check everything is working together.'''
    # To ensure mocks are applied, we import these modules here.
    import octopus_sensing.devices.openbci_streaming as openbci_streaming
    import octopus_sensing.devices.shimmer3_streaming as shimmer3_streaming
    from octopus_sensing.device_coordinator import DeviceCoordinator
    from octopus_sensing.common.message_creators import start_message, stop_message, terminate_message
    from octopus_sensing.monitoring_endpoint import MonitoringEndpoint

    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")

    coordinator = DeviceCoordinator()

    openbci = openbci_streaming.OpenBCIStreaming(
        name="eeg", output_path=output_dir)
    coordinator.add_device(openbci)

    shimmer = shimmer3_streaming.Shimmer3Streaming(
        name="shimmer", output_path=output_dir)
    coordinator.add_device(shimmer)

    monitoring_endpoint = MonitoringEndpoint(coordinator)
    monitoring_endpoint.start()

    try:
        coordinator.dispatch(start_message("int_test", "stimulus_1"))
        # Allowing data collection for five seconds
        time.sleep(5)
        coordinator.dispatch(stop_message("int_test", "stimulus_1"))

        http_client = http.client.HTTPConnection("127.0.0.1:9330")
        http_client.request("GET", "/")
        response = http_client.getresponse()
        assert response.status == 200
        monitoring_data = pickle.loads(response.read())
        assert isinstance(monitoring_data, dict)

        assert isinstance(monitoring_data["eeg"], list)
        # three seconds * data rate
        assert len(monitoring_data["eeg"]) == 3 * 128
        assert len(monitoring_data["eeg"][0]) in (34, 35)
        assert len(monitoring_data["eeg"][-1]) in (34, 35)

        assert isinstance(monitoring_data["shimmer"], list)
        assert len(monitoring_data["shimmer"]) == 3 * 128
        assert len(monitoring_data["shimmer"][0]) in (8, 9)
        assert len(monitoring_data["shimmer"][-1]) in (8, 9)

    finally:
        coordinator.dispatch(terminate_message())
        monitoring_endpoint.stop()

    # To ensure termination is happened.
    time.sleep(0.5)

    eeg_output = os.path.join(output_dir, "eeg")
    assert os.path.exists(eeg_output)
    assert len(os.listdir(eeg_output)) == 1
    assert os.listdir(eeg_output)[0] == "eeg-int_test.csv"

    shimmer_output = os.path.join(output_dir, "shimmer")
    assert os.path.exists(shimmer_output)
    assert len(os.listdir(shimmer_output)) == 1
    assert os.listdir(shimmer_output)[0] == "shimmer-int_test.csv"



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import gi
gi.require_version('Gtk', '3.0')  # nopep8
from gi.repository import Gtk  # nopep8


class Question:
    '''
    The base class for creating questions using Gtk 3.0

    Attributes
    ----------

    Parameters
    ----------
    id: str
        A unique ID for the question
    
    text: str
        The text of question
    '''
    def __init__(self, id: str, text: str):
        self._text = text
        self.id = id

    def render(self, grid: Gtk.Grid, grid_row: int):
        raise NotImplementedError()

    def get_answer(self):
        raise NotImplementedError()



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from screeninfo import get_monitors
from gi.repository import Gtk, GdkPixbuf, GLib, Gst
import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Gst', '1.0')


Gst.init(None)
Gst.init_check(None)


class ImageWindow(Gtk.Window):
    '''
    Creates a Gtk window with a message for informing the participant about something
    It has a continue button which by clicking on it, the window will be destroyed

    Attributes
    ----------

    Parameters
    ----------

    image_path: str
        The path of image
    
    timeout: int
        The time period for displaying the image
    
    monitor_no: int, default: 0
        The ID of monitor for displaying of image. It can be 0, 1, ...

    '''
    def __init__(self, image_path, timeout, monitor_no=0):
        Gtk.Window.__init__(self, title="")

        self._timeout = timeout
        image_box = Gtk.Box()
        monitors = get_monitors()
        image_width = monitors[monitor_no].width
        image_height = monitors[monitor_no].height
        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(
            image_path, image_width, image_height, False)
        image = Gtk.Image()
        image.set_from_pixbuf(pixbuf)
        image_box.pack_start(image, False, False, 0)
        self.add(image_box)

        self.modal = True
        self.fullscreen()
        
        image_box.show()
        image.show()

    def show_window(self):
        GLib.timeout_add_seconds(self._timeout, self.destroy)
        self.show()




# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Foobar.
# If not, see <https://www.gnu.org/licenses/>.

from octopus_sensing.common.message import Message


class MessageType():
    '''
    Predefined message types including:

        1- START: start of a stimulus

        2- STOP: End of each stimulus

        3- TERMINATE: terminating the process of each device
    '''
    START = "START"
    STOP = "STOP"
    TERMINATE = "TERMINATE"


def start_message(experiment_id: str, stimulus_id: str, payload=None):
    '''
    Creates a message to inform device of starting the stimulus

    Parameters
    -----------
    experiment_id: str, default: None
        A unique ID for each participant and task in the study
    
    stimulus_id: str, default: None
        A unique ID for each stimulus

    payload: Any, default: None
        the message data that can have differnt values
    

    Returns
    -------
    message: Message
        A start message

    Example
    --------
    In this example, we created a stop message. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices that stimulus `stimulus_00` has started.

    >>> message = start_message("study_1_p10", "stimulus_00")
    >>> device_coordinator.dispatch(message)

    '''
    return Message(MessageType.START,
                   payload,
                   experiment_id=experiment_id,
                   stimulus_id=stimulus_id)


def stop_message(experiment_id: str, stimulus_id: str):
    '''
    Creates a message to inform device of stopping the stimulus


    Parameters
    -----------
    experiment_id: str, default: None
        A unique ID for each participant and task in the study
    
    stimulus_id: str, default: None
        A unique ID for each stimulus
    

    Returns
    -------
    message: Message
        A start message

    Example
    --------
    In this example, we created a start message. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices that stimulus `stimulus_00` is finished.

    >>> message = stop_message("study_1_p10", "stimulus_00")
    >>> device_coordinator.dispatch(message)

    '''
    return Message(MessageType.STOP,
                   None,
                   experiment_id=experiment_id,
                   stimulus_id=stimulus_id)


def terminate_message():
    '''
    Creates a message to inform device of terminating the program
    

    Returns
    -------
    message: Message
        A start message

    Example
    --------
    In this example, we created a terminate message. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices to terminate.

    >>> message = terminate()
    >>> device_coordinator.dispatch(message)

    '''

    return Message(MessageType.TERMINATE,
                   None)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from screeninfo import get_monitors
from gi.repository import Gtk, GdkPixbuf, GLib, Gst, GObject
import datetime
import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Gst', '1.0')

FONT_STYLE = "<span font_desc='Tahoma {0}'>{1}</span>"
# Todo change it to config
monitors = get_monitors()
image_width = monitors[0].width
image_height = monitors[0].height


class TimerWindow(Gtk.Window):
    '''
    Creates a Gtk window with a timer
    It is like a button which by clicking on it, the timer window will be destroyed

    Attributes
    ----------

    Parameters
    ----------

    title: str
        Title of window
    
    message: str:
        A messageto be displayed on the window

    width: int, default: 400
        The width of questionnaire window in pixel
    
    height: int, default: 200
        The height of questionnaire window in pixel
    
    font_size: str
        The font size for displaying the text

    '''
    def __init__(self, title:str, message: str="", width:int= 400, height: int= 200, font_size=20):
        self._destroy = False
        Gtk.Window.__init__(self, title=title)
        self.set_default_size(width, height)
        grid = Gtk.Grid(column_homogeneous=False,
                        column_spacing=30,
                        row_spacing=30)
        self._font_size = font_size
        label = Gtk.Label()
        label.set_markup(FONT_STYLE.format(self._font_size, message))
        grid.attach(label, 0, 0, 1, 1)
        self._timer_button = Gtk.Button.new_with_label("")
        self._timer_button.connect("clicked", self._on_click_timer_button)
        self._timer_button.get_child().set_markup(FONT_STYLE.format(self._font_size, "0 : 0"))
        Gtk.Widget.set_size_request(self._timer_button, 600, 300)
        grid.attach(self._timer_button, 0, 1, 1, 1)
        self.add(grid)
        self.__min = 0
        self.__sec = 0
        self.set_keep_above(True)

    # Displays Timer
    def _display_timer(self, *args):
        if self._destroy:
            return False

        self.__sec += 1
        if self.__sec >= 60:
            self.__sec = 0
            self.__min += 1
        now_time = str(self.__min) + " : " + str(self.__sec)
        self._timer_button.get_child().set_markup(
            FONT_STYLE.format(self._font_size, now_time))
        return True

    # Initialize Timer
    def _start_timer(self):
        #  this takes 2 args: (how often to update in millisec, the method to run)
        GObject.timeout_add(1000, self._display_timer)

    def show_window(self):
        '''
        Shwos timer window
        '''
        self._start_timer()
        self.show_all()

    def _on_click_timer_button(self, button: Gtk.Button):
        '''
        Destroy the timer window

        Parameters
        ----------
        button: Gtk.Button
        '''
        print("destroy")
        self._destroy = True
        self.destroy()



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from octopus_sensing.common.endpoint_base import EndpointBase


class MonitoringEndpoint(EndpointBase):

    def __init__(self, device_coordinator):
        super().__init__(endpoint_name="MonitoringEndpoint-Thread",
                         port=9330, get_callback=self._get_handler)
        self._device_coordinator = device_coordinator

    def _get_handler(self, request_reader):
        return self._device_coordinator.get_monitoring_data()



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os
import array
from typing import List
import miniaudio

from octopus_sensing.devices.device import Device
from octopus_sensing.common.message_creators import MessageType

class AudioStreaming(Device):
    '''
    Stream and Record audio

    Attributes
    ----------

    Parameters
    ----------

    device_id: int
        The audio recorder ID. If there is several audio recorder in the system
    
    name: str, optional
          device name
          This name will be used in the output path to identify each device's data
    
    output_path: str, optional
                 The path for recording files.
                 Audio files will be recorded in folder {output_path}/{name}


    Example
    -------
    If you want to know what is your audio recorder's ID run the following example  from `miniaudio <https://github.com/irmen/pyminiaudio>`_

    >>> import miniaudio
    >>> devices = miniaudio.Devices()
    >>> captures = devices.get_captures()
    >>> for d in enumerate(captures):
            print("{num} = {name}".format(num=d[0], name=d[1]['name']))

    Example
    -----------
    Creating an instance of audio recorder and adding it to the device coordinator.
    Device coordinator is responsible for triggerng the audio recorder to start or stop recording

    >>> audio_recorder = AudioStreaming(1,
    ...                                 name="Audio_monitoring",
    ...                                 output_path="./output")
    >>> device_coordinator.add_device(audio_recorder)

    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`

    '''
    def __init__(self, device_id:int, **kwargs):
        super().__init__(**kwargs)
        self.output_path = os.path.join(self.output_path, self.name)
        os.makedirs(self.output_path, exist_ok=True)

        self._device_id = device_id
        
        self._stream_data: List[bytes] = []
        self._record = False
        self._terminate = False
        self._state = ""

    def __stream_loop(self):
        _ = yield
        while True:
            data = yield
            self._stream_data.append(data)

    def _run(self):
        devices = miniaudio.Devices()
        captures = devices.get_captures()
        selected_device = captures[self._device_id]
        capture = \
            miniaudio.CaptureDevice(buffersize_msec=1000,
                                    sample_rate=44100,
                                    device_id=selected_device["id"])

        recorder = self.__stream_loop()
        next(recorder)

        while True:
            message = self.message_queue.get()
            if message is None:
                continue
            if message.type == MessageType.START:
                if self._state == "START":
                    print("Audio streaming has already started")
                else:
                    self._stream_data = []
                    capture.start(recorder)
                    self._record = True
                    self._state = "START"
            elif message.type == MessageType.STOP:
                if self._state == "STOP":
                   print("Audio streaming has already stopped")
                else:
                    capture.stop()
                    self._record = False
                    file_name = \
                        "{0}/{1}-{2}-{3}.wav".format(self.output_path,
                                                    self.name,
                                                    message.experiment_id,
                                                    str(message.stimulus_id).zfill(2))
                    self._save_to_file(file_name, capture)
                    self._state = "STOP"
            elif message.type == MessageType.TERMINATE:
                self._terminate = True
                break

    def _save_to_file(self, file_name:str, capture:miniaudio.CaptureDevice):
        buffer = b"".join(self._stream_data)
        samples = array.array('h')
        samples.frombytes(buffer)
        sound = miniaudio.DecodedSoundFile('capture',
                                           capture.nchannels,
                                           capture.sample_rate,
                                           capture.format,
                                           samples)
        miniaudio.wav_write_file(file_name, sound)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os

from typing import List
import csv
import gi
gi.require_version('Gtk', '3.0')  # nopep8
from gi.repository import Gtk  # nopep8

from octopus_sensing.questionnaire.question import Question  # nopep8


class Questionnaire(Gtk.Window):
    '''
    Creating a questionnaire using Gtk
    It has a Done button which by clicking on it, the answers will be saved and window will be destroyed

    Attributes
    -----------

    Parameters
    ----------
    name: str
        Questionnaire name
    
    experiment_id: str
        The unique ID of the experiment

    stimulus_id: str
        The unique ID of the stimulus
    
    title: str
        Questionnaire title
    
    width: int, default: 400
        The width of questionnaire window in pixel
    
    height: int, default: 200
        The height of questionnaire window in pixel
    
    output_path: str
        The output path for recording answers
    
    Examples
    --------
    Creating an opinion question, a text question and adding it to the questionnaire

    >>> emotions = {"Happiness": 4, "Sadness": 6, "Neutral": 5, "Fear": 3, "Anger": 1}
    >>> question_1 = OpinionQuestion("q1",
    ...                              "1- What emotion did you feel the most?",
    ...                              options=emotions,
    ...                              default_answer=5)
    >>> question_2 = TextQuestion("q2",
    ...                           "1- Was your feelings positive or negative?",
    ...                           default_answer="Positive")
    >>> questionnaire = Questionnaire("after_stimuli",
    ...                               "study01_p10",
    ...                               "stimuli00",
    ...                               "After Stimulus Questionnaire")
    >>> questionnaire.add_questions([question_1, question_2])

    See Also
    -----------
    :class:`octopus_sensing.questionnaire.text_question`
    :class:`octopus_sensing.questionnaire.opinion_question`

    '''
    def __init__(self, name: str, experiment_id: str, stimulus_id: str,
                 title: str, width: int = 400, height: int = 200, output_path="output/self_report"):
        Gtk.Window.__init__(self, title=title)
        self.set_border_width(10)
        self.set_default_size(width, height)
        self._name = name
        self._questions: List[Question] = []
        self._experiment_id = experiment_id
        self.stimulus_id = stimulus_id
        self._output_path = output_path
        if not os.path.exists(self._output_path):
            os.mkdir(self._output_path)

    def add_question(self, question: Question) -> None:
        '''
        Adds a question to the questionnaire

        Parameters
        -----------
        question Question: a question
        '''
        assert isinstance(question, Question)
        if question.id in self._questions:
            raise RuntimeError(
                "The question ID {0} already exists in the questionnaire".format(question.id))
        self._questions.append(question)

    def add_questions(self, questions: List[Question]) -> None:
        '''
        Adds a list of questions to the questionnaire

        Parameters
        ----------
        questions: List(Question)
            a list of questions

        '''
        for question in questions:
            self.add_question(question)

    def show(self) -> None:
        '''
        Shows the questionnaire
        '''
        grid = Gtk.Grid(column_homogeneous=False,
                        column_spacing=10,
                        row_spacing=10)
        self.add(grid)
        i = 0
        for question in self._questions:
            row = question.render(grid, i)
            i += row
        done_button = Gtk.Button.new_with_label("Done")
        done_button.connect("clicked", self._on_click_done_button)
        done_button.get_child().set_markup("<span font_desc='Tahoma 14'>Done</span>")
        grid.attach(done_button, 0, i, 1, 1)
        self.connect("destroy", Gtk.main_quit)
        self.show_all()
        Gtk.main()

    def _on_click_done_button(self, button: Gtk.Button) -> None:
        '''
        Saves answers and close the questionnaire

        Parameters
        ----------
        button: Gtk.Button
            by clicking this button, this method will call

        '''
        self._save_answers()
        self.destroy()

    def _save_answers(self) -> None:
        '''
        Saves answers
        '''
        file_name = \
            "{0}/{1}-{2}.csv".format(self._output_path,
                                     self._name,
                                     self._experiment_id)
        if not os.path.exists(file_name):
            csv_file = open(file_name, 'a')
            header = ["stimulus ID"]
            for question in self._questions:
                header.append(question.id)
            writer = csv.writer(csv_file)
            writer.writerow(header)
            csv_file.flush()
            csv_file.close()

        row = [self.stimulus_id]
        for question in self._questions:
            row.append(question.get_answer())

        with open(file_name, 'a') as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(row)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import gi
gi.require_version('Gtk', '3.0')  # nopep8
from gi.repository import Gtk  # nopep8

FONT_STYLE = "<span font_desc='Tahoma 18'>{}</span>"

class MessageWindow(Gtk.Window):

    '''
    Creating a message window using Gtk
    It has a button which by clicking on it, The window will be closed

    Attributes
    -----------

    Parameters
    ----------
    title: str
        Window title
    
    message: str
        The message text
    
    width: int, default: 500
        The width of questionnaire window in pixel
    
    height: int, default: 200
        The height of questionnaire window in pixel

    '''
    def __init__(self, title: str, message: str, button_label: str = "Ok",
                 width: int = 500, height: int = 200):
        Gtk.Window.__init__(self, title=title)
        self.set_border_width(10)
        self.set_default_size(width, height)
        self._message = message
        self._width = width
        self._height = height
        self._button_label = button_label


    def show(self) -> None:
        '''
        Shows the message
        '''
        grid = Gtk.Grid(column_homogeneous=False,
                        column_spacing=10,
                        row_spacing=10)
        self.add(grid)

        message_label_box = Gtk.Box(spacing=120)
        message_label = Gtk.Label()
        message_label.set_markup(FONT_STYLE.format(self._message))
        Gtk.Widget.set_size_request(message_label, self._width, (self._height - 50))
        message_label_box.pack_start(message_label, False, False, 0)

        grid.attach(message_label_box, 0, 0, 1, 1)


        ok_button = Gtk.Button.new_with_label("Ok")
        ok_button.connect("clicked", self._on_click_ok_button)
        ok_button.get_child().set_markup("<span font_desc='Tahoma 14'>{}</span>".format(self._button_label))
        Gtk.Widget.set_size_request(ok_button, self._width, 50)
        grid.attach(ok_button, 0, 1, 1, 1)
        self.connect("destroy", Gtk.main_quit)
        self.show_all()
        Gtk.main()

    def _on_click_ok_button(self, button: Gtk.Button) -> None:
        '''
        Close the message dialog

        Parameters
        ----------
        button: Gtk.Button
            by clicking this button, this method will be called

        '''
        self.destroy()


# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.
import os
from typing import List, Optional
import pandas as pd
import numpy as np
import mne

from octopus_sensing.preprocessing.utils import load_all_trials, resample, load_all_samples
from octopus_sensing.devices.common import SavingModeEnum


def openbci_preprocess(input_path: str, file_name: str, output_path: str,
                       channels: List[str],
                       saving_mode: int = SavingModeEnum.CONTINIOUS_SAVING_MODE,
                       sampling_rate: int = 128,
                       signal_preprocess: bool = True):
    '''
    Preprocess openbci recorded files to prepare them for visualizing and analysis
    It applys data cleaning (according to signal_preprocess), resampling (according to sampling_rate),
    and splits data if data has been recorded continuously.
    This method uses `mne library <https://mne.tools/stable/index.html>`_ for EEG data processing

    Parameters
    ----------
    input_path: str
        The path to recorded openbci data
    
    file_name: str
        The file name of recorded openbci data
    
    output_path: str
        preprocessed file path
    
    channels: List(str)
        a list of recorded channels name
    
    saving_mode: int, default: SavingModeEnum.CONTINIOUS_SAVING_MODE
        The saving mode of recorded data. If it is CONTINIOUS_SAVING_MODE, data will be splitted
        according to markers and will be recorded in the separated files

    sampling_rate: int, default: 128
        The desired sampling_rate. Data will be resampled according to this sampling rate
    
    signal_preprocess: bool, default: True
        If True will apply preliminary preprocessing steps to clean line noises
    
    Note
    -----
    Sometimes recorded data in one second with Openbci are less or more than 
    the specified sampling rate. So, we resample data by replicating
    the last samples or removing some samples to achieve the desired sampling_rate
    '''
    if saving_mode == SavingModeEnum.SEPARATED_SAVING_MODE:
        if len(channels) == 8:
            data, times = \
                load_all_samples(os.path.join(input_path, file_name),
                                 (0, 8),  # channel columns
                                 12,
                                 '%H:%M:%S.%f')  # timestamp column

        elif len(channels) == 16:
            data, times = \
                load_all_samples(os.path.join(input_path, file_name),
                                 (0, 16),  # channel columns
                                 20,
                                 '%H:%M:%S.%f')  # timestamp column
        output_file_path = \
            "{0}/{1}".format(output_path, file_name)
        resampled_data = \
            resample(data, times, sampling_rate)
        if signal_preprocess is True:
            preprocessed_data = \
                clean_eeg(resampled_data/1e6,
                          channel_names=channels,
                          sampling_rate=sampling_rate)
        else:
            preprocessed_data = resampled_data
            # convert array into dataframe
        data_frame = pd.DataFrame(preprocessed_data, columns=channels)
        data_frame.to_csv(output_file_path, index=False)

    elif saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
        if len(channels) == 8:
            trials_data, trials_times, triger_list = \
                load_all_trials(os.path.join(input_path, file_name),  # File path
                                (0, 8),  # channel columns
                                12,  # timestamp column
                                13,  # triger column
                                '%H:%M:%S.%f')  # timestamp format
        elif len(channels) == 16:
            trials_data, trials_times, triger_list = \
                load_all_trials(os.path.join(input_path, file_name),  # File path
                                (0, 16),  # channel columns
                                20,  # timestamp column
                                21,  # triger column
                                '%H:%M:%S.%f')  # timestamp format
        print("len trials ***************", len(trials_data))
        i = 0
        for trial in trials_data:
            output_file_path = \
                "{0}/{1}-{2}.csv".format(output_path,
                                         file_name[:-4],  # Removing .csv from file_name
                                         str(triger_list[i]).zfill(2))
            print("output_file_path", output_file_path)
            resampled_data = \
                resample(trial, trials_times[i], sampling_rate)

            if signal_preprocess is True:
                preprocessed_data = \
                    clean_eeg(resampled_data,
                              channel_names=channels,
                              sampling_rate=sampling_rate)
            else:
                preprocessed_data = resampled_data
                # convert array into dataframe
            data_frame = pd.DataFrame(preprocessed_data, columns=channels)

            # save the dataframe as a csv file
            data_frame.to_csv(output_file_path, index=False)
            i += 1
    else:
        raise Exception("Saving mode is incorrect")


class EegPreprocessing():
    '''
    Converts EEG data to mne raw data format for furthur analysis

    Parameters
    ----------
    data: numpy.ndarray
        The channels’ time series (n_samples*n_channels)
    
    channel_names: List[str], default: None
        A list of channels' names
    
    sampling_rate: int, default: 128
        Sampling rate of data

    '''
    def __init__(self, data: np.ndarray, channel_names: List[str]=None, sampling_rate: int=128):
        if channel_names is None:
            self._channel_names = \
                ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3",
                 "C4", "T4", "T5", "P3", "P4", "T6", "O1", "O2"]
        else:
            self._channel_names = channel_names

        channel_data = np.transpose(data)
        channel_types = ["eeg"]*len(self._channel_names)

        self.__info = mne.create_info(self._channel_names,
                                      sampling_rate,
                                      channel_types)
        montage = mne.channels.make_standard_montage('standard_1020')
        self._mne_raw = mne.io.RawArray(channel_data, self.__info)
        self._mne_raw.set_montage(montage, match_case=False)
        self._mne_raw.load_data()

    def get_data(self):
        '''
        Gets EEG data as a raw mne data

        Returns
        --------
        mne_raw_data: mne.io.RawArray
            EEG data as a  mne.io.RawArray

        '''
        return self._mne_raw.get_data()

    def filter_data(self, low_frequency: float=1, high_frequency:float =45, notch_frequencies: List[int]=[60]):
        '''
        Apply notch filter, low pass and high pass (bandpass) filter on mne data

        Parameters
        -----------
        low_frequency: float, default: 1
            The low cut frequency for filtering
        
        high_frequency: float, default: 45
            The high cut frequency for filtering
        
        notch_frequencies: List[int] default: [60]
            the frequencies to be used in the notch filter

        '''
        # Band pass filter
        self._mne_raw.filter(l_freq=low_frequency, h_freq=high_frequency)
        # Notch filter
        if notch_frequencies not in (None, []):
            self._mne_raw.notch_filter(notch_frequencies)


def clean_eeg(data, channel_names: List[str] = None,
              low_frequency: float = 1,
              high_frequency: float = 45,
              sampling_rate: int = 128):
    '''
    Cleans EEG data

    Parameters
    -----------
    low_frequency: float, default: 1
        The low cut frequency for filtering
    
    high_frequency: float, default: 45
        The high cut frequency for filtering
    
    smpling_rate: int, default: 128
        sampling rate
    
    '''
    if channel_names is None:
        channel_names = ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3",
                         "C4", "T4", "T5", "P3", "P4", "T6", "O1", "O2"]

    # The length of data is 5 minutes
    preprocessing = \
        EegPreprocessing(data,
                         channel_names=channel_names,
                         sampling_rate=sampling_rate)
    preprocessing.filter_data(low_frequency=low_frequency, high_frequency=high_frequency)
    preprocessed_data = preprocessing.get_data()
    return np.transpose(preprocessed_data)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os
import threading
import csv
import datetime
import pyOpenBCI
import numpy as np

from octopus_sensing.common.message_creators import MessageType
from octopus_sensing.devices.monitored_device import MonitoredDevice
from octopus_sensing.devices.common import SavingModeEnum


uVolts_per_count = (4500000)/24/(2**23-1)
accel_G_per_count = 0.002 / (2**4)  # G/count


class OpenBCIStreaming(MonitoredDevice):
    '''
    Manages OpenBCI streaming
    It uses pyOpenBCI library which is not supporting by OpenBCI anymore.
    We recommend to use 
    :class:`octopus_sensing.devices.brainflow_openbci_streaming` instead.

    Data will be recorded in a csv file/files with the following column order:
    channels, Acc_x, Acc_y, Acc_z, sample_id, time_stamp, trigger

    Attributes
    -----------

    Parameters
    ----------
    name: str, default: None
        Device name. This name will be used in the output path to identify 
        each device's data.

    output_path: str,  default: output
        The path for recording files.
        Audio files will be recorded in folder {output_path}/{name}
    
    saving_mode: int, default: SavingModeEnum.CONTINIOUS_SAVING_MODE
        The way of saving data: saving continiously in a file or save data related to
        each stimulus in a separate file. 
        SavingModeEnum is:

            0. CONTINIOUS_SAVING_MODE
            1. SEPARATED_SAVING_MODE
    
    daisy: bool, default: True
           If it is True, it means we use cyton-daisy board,
           otherwise we use cyton board

    channels_order: list of str, default: None
        A list of channel names which specify the order and names of channels

    Example
    --------
    Creating an instance of OpenBCI board with USB dongle using 
    `pyOpenBCI <https://github.com/andreaortuno/pyOpenBCI>`_,
    and adding it to the device coordinator. Device coordinator is responsible
    for triggerng the OpenBCI to start or stop recording  or to add markers to
    recorded data.
    In this example, since the saving mode is continuous, all recorded data 
    will be saved in a file. But, when an event happens, device coordinator will send
    a trigger message to the device and recorded data will be marked with the trigger

    >>> my_openbci = 
    ...     OpenBCIStreaming(name="OpenBCI",
    ...                      output_path="./output",
    ...                      daisy=True,
    ...                      saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE,
    ...                      channels_order=["Fp1", "Fp2", "F7", "F3", 
    ...                                      "F4", "F8", "T3", "C3",
    ...                                      "C4", "T4", "T5", "P3", 
    ...                                      "P4", "T6", "O1", "O2"])
    >>> device_coordinator.add_device(my_openbci)

    Note
    -----
    Before running the code, turn on the OpenBCI, connect the dongle and make sure its port is free.


    See Also
    --------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`,
    :class:`octopus_sensing.devices.brainflow_openbci_streaming`

    '''
    def __init__(self,
                 daisy=True,
                 channels_order=None,
                 saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                 **kwargs):
        super().__init__(**kwargs)

        self._saving_mode = saving_mode
        self._stream_data = []
        self._board = self._inintialize_board(daisy)
        self._trigger = None
        self._experiment_id = None

        self.output_path = self._make_output_path()

        if channels_order is None:
            if daisy is True:
                self.channels = \
                    ["ch1", "ch2", "ch3", "ch4", "ch5", "ch6", "ch7", "ch8", "ch9",
                     "ch10", "ch11", "ch12", "ch13", "ch14", "ch15", "ch16"]
            else:
                self.channels = \
                    ["ch1", "ch2", "ch3", "ch4", "ch5", "ch6", "ch7", "ch8"]
        else:
            self.channels = channels_order
            if daisy is True:
                if len(self.channels) != 16:
                    raise "The number of channels in channels_order should be 16"
            elif daisy is False:
                if len(self.channels) != 8:
                    raise "The number of channels in channels_order should be 8"

    def _make_output_path(self):
        output_path = os.path.join(self.output_path, "eeg")
        os.makedirs(output_path, exist_ok=True)
        return output_path

    def _inintialize_board(self, daisy):
        return pyOpenBCI.OpenBCICyton(daisy=daisy)

    def _run(self):
        threading.Thread(target=self._stream_loop).start()

        while True:
            message = self.message_queue.get()
            if message is None:
                continue
            if message.type == MessageType.START:
                self.__set_trigger(message)
                self._experiment_id = message.experiment_id
            elif message.type == MessageType.STOP:
                if self._saving_mode == SavingModeEnum.SEPARATED_SAVING_MODE:
                    self._experiment_id = message.experiment_id
                    file_name = \
                        "{0}/{1}-{2}-{3}.csv".format(self.output_path,
                                                     self.name,
                                                     self._experiment_id,
                                                     message.stimulus_id)
                    self._save_to_file(file_name)
                else:
                    self._experiment_id = message.experiment_id
                    self.__set_trigger(message)
            elif message.type == MessageType.TERMINATE:
                if self._saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
                    file_name = \
                        "{0}/{1}-{2}.csv".format(self.output_path,
                                                 self.name,
                                                 self._experiment_id)
                    self._save_to_file(file_name)
                break

        self._board.stop_stream()

    def __set_trigger(self, message):
        '''
        Takes a message and set the trigger using its data

        @param Message message: a message object
        '''
        self._trigger = \
            "{0}-{1}-{2}".format(message.type,
                                 message.experiment_id,
                                 str(message.stimulus_id).zfill(2))

    def _stream_loop(self):
        self._board.start_stream(self._stream_callback)

    def _stream_callback(self, sample):
        data = np.array(sample.channels_data) * uVolts_per_count
        acc_data = np.array(sample.aux_data) * accel_G_per_count
        data_list = list(data) + list(acc_data)
        data_list.append(sample.id)
        data_list.append(str(datetime.datetime.now().time()))
        if self._trigger is not None:
            data_list.append(self._trigger)
            self._trigger = None
        self._stream_data.append(data_list)

    def _save_to_file(self, file_name):
        if not os.path.exists(file_name):
            csv_file = open(file_name, 'a')
            header = []
            header.extend(self.channels)
            header.extend(["acc-x", "acc-y", "acc-z"])
            header.extend(["sample_id", "time stamp", "trigger"])
            writer = csv.writer(csv_file)
            writer.writerow(header)
            csv_file.flush()
            csv_file.close()
        with open(file_name, 'a') as csv_file:
            writer = csv.writer(csv_file)
            for row in self._stream_data:
                writer.writerow(row)
                csv_file.flush()

    def _get_monitoring_data(self):
        '''Returns latest collected data for monitoring/visualizing purposes.'''
        # Last three seconds
        # FIXME: hard-coded data collection rate
        return self._stream_data[-1 * 3 * 128:]

    def get_saving_mode(self):
        '''
        Gets saving mode
        
        Returns
        -----------
        saving_mode: int
            The way of saving data: saving continiously in a file or save data related to
            each stimulus in a separate file. 
            SavingModeEnum is CONTINIOUS_SAVING_MODE = 0 or SEPARATED_SAVING_MODE = 1
        '''
        return self._saving_mode

    def get_output_path(self):
        '''
        Gets the path that is used for data recording

        Returns
        -----------
        output_path: str
           The output path that use for data recording
        '''
        return self.output_path

    def get_channels(self):
        '''
        Gets the list of channels

        Returns
        -------

        channels_name: List[str]
            The list of channels' name

        '''
        return self.channels



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Foobar.
# If not, see <https://www.gnu.org/licenses/>.

import time
import pytest

from octopus_sensing.device_coordinator import MonitoringCache, DeviceCoordinator
from octopus_sensing.devices.device import Device


def test_monitoring_cache():
    cache = MonitoringCache()

    data = [1, 2, 3]

    cache.cache(data)

    time.sleep(0.02)
    assert cache.get_cache() == data

    time.sleep(0.1)
    assert cache.get_cache() is None


def fake_run():
    pass

def test_should_auto_assign_device_id():
    test_device = Device()
    test_device._run = fake_run
    coordinator = DeviceCoordinator()
    coordinator.add_device(test_device)
    assert isinstance(test_device.name, str)
    assert len(test_device.name) > 0


def test_should_not_add_duplicated_deivces():
    test_device = Device(name="device1")
    test_device._run = fake_run
    coordinator = DeviceCoordinator()
    coordinator.add_device(test_device)
    with pytest.raises(RuntimeError):
        coordinator.add_device(test_device)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os
import platform
import threading
import datetime
import csv
import math
import struct
import serial
from typing import List, Optional

from octopus_sensing.devices.monitored_device import MonitoredDevice
from octopus_sensing.common.message_creators import MessageType
from octopus_sensing.common.message import Message
from octopus_sensing.devices.common import SavingModeEnum

class Shimmer3Streaming(MonitoredDevice):
    '''
    Streams and Records Shimmer3 data.
    Data will be recorded in a csv file/files with the following column order:
    type, time stamp, Acc_x, Acc_y, Acc_z, GSR_ohm, PPG_mv, time, trigger

    Attributes
    ----------

    Parameters
    ----------
    name: str, default: None
        Device name. This name will be used in the output path to identify 
        each device's data.

    output_path: str,  default: output
        The path for recording files.
        Audio files will be recorded in folder {output_path}/{name}
    
    saving_mode: int, default: SavingModeEnum.CONTINIOUS_SAVING_MODE
        The way of saving data: saving continiously in a file or save data related to
        each stimulus in a separate file. 
        SavingModeEnum is:
        
            0. CONTINIOUS_SAVING_MODE
            1. SEPARATED_SAVING_MODE
    
    sampling_rate: int, default: 128
        The sampling frequency for acquiring data from the device
    

    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`

    Example
    -------
    Creating an instance of shimmer3 and adding it to the device coordinator.
    Device coordinator is responsible for triggerng the shimmer3 to 
    start or stop recording  or to add markers to recorded data.
    In this example, since the saving mode is continuous, all recorded data 
    will be saved in a file. But, when an event happens, device coordinator will send a trigger message
    to the device and recorded data will be marked with the trigger

    >>> my_shimmer = Shimmer3Streaming(name="shimmer",
    ...                                saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE,
    ...                                output_path="./output")
    >>> device_coordinator.add_device(my_shimmer)


    Note
    -----
    Keep in your mind, before running the code for Shimmer data recording,
    turn on the Shimmer3 sensor and pair bluetooth and the serial port. (Shimmer password: 1234)

    For example in linux you can do it as follow:
        1- hcitool scan   //It shows the macaddress of device. for shimmer it is 00:06:66:F0:95:95

        2- vim /etc/bluetooth/rfcomm.conf write the below line in it: 
        rfcomm0{ bind no; device 00:06:66:F0:95:95; channel 1; comment "serial port" } 

        3- sudo rfcomm connect rfcomm0 00:06:66:F0:95:95 // This is for reading bluetooth data from a serial port

    Note
    -----
    This class is based on `ShimmerReader <https://github.com/nastaran62/ShimmerReader>`_ 
    which is an extended version of 
    `LogAndStream python firmware <http://www.shimmersensing.com/images/uploads/docs/LogAndStream_for_Shimmer3_Firmware_User_Manual_rev0.11a.pdf>`_ 
    for Shimmer3 data streaming.
    '''

    def __init__(self,
                 sampling_rate: int=128,
                 saving_mode: int=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                 **kwargs):
        super().__init__(**kwargs)

        self._saving_mode = saving_mode
        self._stream_data: List[float] = []
        self._sampling_rate = sampling_rate
        self._inintialize_connection()
        self._trigger: Optional[str] = None
        self._break_loop = False
        self.output_path = self._make_output_path()
        self._state = ""

    def _make_output_path(self):
        output_path = os.path.join(self.output_path, self.name)
        os.makedirs(output_path, exist_ok=True)
        return output_path

    def _inintialize_connection(self):
        '''
        Initializing connection with Simmer3 device
        '''
        os_type = platform.system()
        if os_type == "Windows":
            self._serial = serial.Serial("Com12", 115200)
        else:
            self._serial = serial.Serial("/dev/rfcomm0", 115200)
        self._serial.flushInput()
        print("port opening, done.")
        # send the set sensors command
        # 4 bytes command:
        #     0x08 is SET_SENSORS_COMMAND
        #     Each bit in the three following bytes are one sensor.
        self._serial.write(struct.pack(
            'BBBB', 0x08, 0x84, 0x01, 0x00))  # GSR and PPG
        self._wait_for_ack()
        print("sensor setting, done.")

        # Enable the internal expansion board power
        self._serial.write(struct.pack('BB', 0x5E, 0x01))
        self._wait_for_ack()
        print("enable internal expansion board power, done.")

        # send the set sampling rate command

        '''
        sampling_freq = 32768 / clock_wait = X Hz
        2 << 14 = 32768
        '''
        clock_wait = math.ceil((2 << 14) / self._sampling_rate)

        self._serial.write(struct.pack('<BH', 0x05, clock_wait))
        self._wait_for_ack()

        # Inquiry configurations (For finding channels order)
        # Page 16 of This PDF:
        # http://www.shimmersensing.com/images/uploads/docs/LogAndStream_for_Shimmer3_Firmware_User_Manual_rev0.11a.pdf

        self._serial.write(struct.pack('B', 0x01))
        self._wait_for_ack()
        inquiery_response = bytes("", 'utf-8')
        # response_size is 1 packet_type + 2 Sampling rate + 4 Config Bytes +
        # 1 Num Channels + 1 Buffer size
        response_size = 9
        numbytes = 0
        while numbytes < response_size:
            inquiery_response += self._serial.read(response_size)
            numbytes = len(inquiery_response)

        num_channels = inquiery_response[7]
        print("Number of Channels:", num_channels)
        print("Buffer size:", inquiery_response[8])

        # There's one byte for each channel
        # For the meaning of each byte, refer to the above PDF
        channels = bytes("", "utf-8")
        numbytes = 0
        while numbytes < num_channels:
            channels += self._serial.read(num_channels)
            numbytes = len(channels)

        print("Channel 1:", channels[0])
        print("Channel 2:", channels[1])
        print("Channel 3:", channels[2])
        print("Channel 4:", channels[3])
        print("Channel 5:", channels[4])

        # send start streaming command
        self._serial.write(struct.pack('B', 0x07))
        self._wait_for_ack()
        print("start command sending, done.")
        self._experiment_id = 0

    def _wait_for_ack(self):
        ddata = ""
        ack = struct.pack('B', 0xff)
        while ddata != ack:
            ddata = self._serial.read(1)

    def _run(self):
        '''
        Listening to the message queue and manage messages
        '''
        loop_thread = threading.Thread(target=self._stream_loop)
        loop_thread.start()

        while True:
            message = self.message_queue.get()
            if message is None:
                continue
            if message.type == MessageType.START:
                if self._state == "START":
                    print("Shimmer3 streaming has already recorded the START triger")
                else:
                    print("Shimmer start")
                    self._experiment_id = message.experiment_id
                    self.__set_trigger(message)
                    self._state = "START"
            elif message.type == MessageType.STOP:
                if self._state == "STOP":
                    print("Shimmer3 streaming has already recorded the STOP triger")
                else:
                    if self._saving_mode == SavingModeEnum.SEPARATED_SAVING_MODE:
                        self._experiment_id = message.experiment_id
                        file_name = \
                            "{0}/{1}-{2}-{3}.csv".format(self.output_path,
                                                        self.name,
                                                        self._experiment_id,
                                                        message.stimulus_id)
                        self._save_to_file(file_name)
                    else:
                        print("Shimmer stop")
                        self._experiment_id = message.experiment_id
                        self.__set_trigger(message)
                    self._state = "STOP"
            elif message.type == MessageType.TERMINATE:
                if self._saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
                    file_name = \
                        "{0}/{1}-{2}.csv".format(self.output_path,
                                                 self.name,
                                                 self._experiment_id)
                    self._save_to_file(file_name)
                break

        self._break_loop = True
        loop_thread.join()

    def __set_trigger(self, message: Message):
        '''
        Takes a message and set the trigger using its data

        Parameters
        -----------
        message: Message
                 a message object
        '''
        self._trigger = \
            "{0}-{1}-{2}".format(message.type,
                                 message.experiment_id,
                                 str(message.stimulus_id).zfill(2))

    def _stream_loop(self):
        '''
        Reads incoming data
        '''
        ddata = bytes("", 'utf-8')
        numbytes = 0
        # 1byte packet type + 3byte timestamp + 2 byte X + 2 byte Y +
        # 2 byte Z + 2 byte PPG + 2 byte GSR
        framesize = 14

        try:
            while True:
                while numbytes < framesize:
                    ddata += self._serial.read(framesize)
                    numbytes = len(ddata)
                    if self._break_loop:
                        break

                if self._break_loop:
                    self._stop_shimmer()
                    break

                data = ddata[0:framesize]
                ddata = ddata[framesize:]
                numbytes = len(ddata)

                # read basic packet information
                (packettype) = struct.unpack('B', data[0:1])
                (timestamp0, timestamp1, timestamp2) = \
                    struct.unpack('BBB', data[1:4])

                # read packet payload
                (x, y, z, PPG_raw, GSR_raw) = \
                    struct.unpack('HHHHH', data[4:framesize])
                record_time = datetime.datetime.now()

                # get current GSR range resistor value
                data_range = ((GSR_raw >> 14) & 0xff)  # upper two bits
                if data_range == 0:
                    rf = 40.2   # kohm
                elif data_range == 1:
                    rf = 287.0  # kohm
                elif data_range == 2:
                    rf = 1000.0  # kohm
                elif data_range == 3:
                    rf = 3300.0  # kohm

                # convert GSR to kohm value
                gsr_to_volts = (GSR_raw & 0x3fff) * (3.0/4095.0)
                GSR_ohm = rf/((gsr_to_volts / 0.5) - 1.0)

                # convert PPG to milliVolt value
                PPG_mv = PPG_raw * (3000.0/4095.0)

                timestamp = timestamp0 + timestamp1*256 + timestamp2*65536

                #print([packettype[0], timestamp, GSR_ohm, PPG_mv] + self._trigger)

                if self._trigger is not None:
                    print("Shimmer trigger")
                    row = [packettype[0],
                           timestamp,
                           x, y, z,
                           GSR_ohm,
                           PPG_mv,
                           record_time,
                           self._trigger]
                    self._trigger = None
                else:
                    row = [packettype[0],
                           timestamp,
                           x, y, z,
                           GSR_ohm,
                           PPG_mv,
                           record_time]
                self._stream_data.append(row)

        except KeyboardInterrupt:
            self._stop_shimmer()

    def _stop_shimmer(self):
        # send stop streaming command
        self._serial.write(struct.pack('B', 0x20))

        print("stop command sent, waiting for ACK_COMMAND")
        self._wait_for_ack()
        print("ACK_COMMAND received.")
        self._serial.close()
        print("All done")

    def _save_to_file(self, file_name):
        if not os.path.exists(file_name):
            csv_file = open(file_name, 'a')
            header = ["type", "time stamp", "Acc_x", "Acc_y", "Acc_z",
                      "GSR_ohm",
                      "PPG_mv",
                      "time",
                      "trigger"]
            writer = csv.writer(csv_file)
            writer.writerow(header)
            csv_file.flush()
            csv_file.close()

        with open(file_name, 'a') as csv_file:
            writer = csv.writer(csv_file)
            for row in self._stream_data:
                writer.writerow(row)
                csv_file.flush()

    def _get_monitoring_data(self):
        '''Returns latest collected data for monitoring/visualizing purposes.'''
        # Last three seconds
        return self._stream_data[-1 * 3 * self._sampling_rate:]

    def get_saving_mode(self):
        '''
        Gets saving mode
        
        Returns
        -----------
        saving_mode: int
            The way of saving data: saving continiously in a file or save data related to
            each stimulus in a separate file. 
            SavingModeEnum is CONTINIOUS_SAVING_MODE = 0 or SEPARATED_SAVING_MODE = 1
        '''
        return self._saving_mode

    def get_output_path(self):
        '''
        Gets the path that is used for data recording

        Returns
        -----------
        output_path: str
           The output path that use for data recording
        '''
        return self.output_path



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Foobar.
# If not, see <https://www.gnu.org/licenses/>.

import time
import pytest
import sys  # nopep8
import threading
import socket

from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.devices.network_devices.socket_device import SocketNetworkDevice
from octopus_sensing.common.message_creators import start_message, stop_message

PORT = 5002
EVENT = threading.Event()
TIMEOUT = 30

def __server():
    '''
    Starts a server and send several messages and terminates
    '''
    device_coordinator = DeviceCoordinator()
    socket_device = SocketNetworkDevice("localhost", PORT)
    device_coordinator.add_devices([socket_device])

    EVENT.wait(timeout=TIMEOUT)
    message = start_message("test", "00")
    device_coordinator.dispatch(message)

    time.sleep(1)
    message = stop_message("test", "00")
    device_coordinator.dispatch(message)
    time.sleep(1)
    message = start_message("test", "01")
    device_coordinator.dispatch(message)
    time.sleep(1)
    message = stop_message("test", "01")
    device_coordinator.dispatch(message)
    time.sleep(2)

    device_coordinator.terminate()

def test_client_socket():
    threading.Thread(target=__server).start()
    time.sleep(5)
    host = "localhost"
    port = PORT

    # Waiting at most 30 seconds for the SocketNetworkDevice to become ready.
    retries = TIMEOUT
    while True:
      try:
          client_socket = socket.create_connection((host, port))
          break
      except ConnectionError:
          if retries <= 0:
              raise
          retries = retries - 1
          time.sleep(1)
          continue
    # Causes the server to start sending messages.
    EVENT.set()

    data = client_socket.recv(1024).decode()  # receive response
    assert data == "START-test-00\n"
    data = client_socket.recv(1024).decode()  # receive response
    assert data == "STOP-test-00\n"
    data = client_socket.recv(1024).decode()  # receive response
    assert data == "START-test-01\n"
    data = client_socket.recv(1024).decode()  # receive response
    assert data == "STOP-test-01\n"

    data = client_socket.recv(1024).decode()  # receive response
    assert data == "terminate\n"

    client_socket.close()  # close the connection



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from octopus_sensing.common.endpoint_base import EndpointBase, EndpointClientError
from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.common.message import Message


class DeviceMessageHTTPEndpoint(EndpointBase):
    '''
    Stars and endpoint that listens for incoming Message requests. It passes the message to
    the Device Coordinator to dispatch them to the devices.
    It accepts HTTP POST requests. The Body can be serialized in one of 'json', 'msgpack'
    or 'pickle'. See the Usage section for how to call the endpoint.

    Attributes
    ----------

    Parameters
    ----------
    device_coordinator
        An instance of DeviceCoordinator class.
    port
        Port to listen on. Default is: 9331

    Examples
    --------
    To start the endpoint:

    >>> from octopus_sensing.device_coordinator import DeviceCoordinator
    >>> from octopus_sensing.device_message_endpoint import DeviceMessageHTTPEndpoint
    >>> device_coordinator = DeviceCoordinator()
    >>> # Add you devices
    >>> message_endpoint = DeviceMessageHTTPEndpoint(device_coordinator)
    >>> message_endpoint.start()
    >>> # You need to stop it after your program finished.
    >>> message_endpoint.stop()

    The client can call the endpoint like this (Note that it's a separate process):

    >>> import http.client
    >>> import msgpack
    >>> http_client = http.client.HTTPConnection("127.0.0.1:9331", timeout=3)
    >>> http_client.request(
    ...     "POST", "/",
    ...    body=msgpack.packb({'type': 'START',
    ...                        'experiment_id': '123',
    ...                        'stimulus_id': 's8'}),
    ...                        headers={'Accept': 'application/msgpack'})
    >>> response = http_client.getresponse()
    >>> assert response.status == 200

    Note that you must pass 'Accept' in the headers, so the endpoint knows what
    type of serialization you used.
    Also, the request needs 'Content-Length' in headers. However, it's part of
    the HTTP standard and your http client should do it automatically.

    '''

    def __init__(self, device_coordinator: DeviceCoordinator, port: int = 9331):
        super().__init__(endpoint_name="DeviceMessageHttpEndpoint-Thread",
                         port=port, post_callback=self._post_handler)
        self._device_coordinator = device_coordinator

    def _post_handler(self, request_body: dict):
        if not isinstance(request_body, dict):
            raise EndpointClientError(
                "The request body must contain a dictionary. Got [{}]".format(type(request_body)))

        if 'type' not in request_body:
            raise EndpointClientError(
                "'type' field is mandatory in request body")

        message_type = request_body['type']
        experiment_id = request_body.get('experiment_id', None)
        stimulus_id = request_body.get('stimulus_id', None)

        if not isinstance(message_type, str):
            raise EndpointClientError(
                "'type' must be of type 'str', got '{0}'".format(type(message_type)))
        if experiment_id is not None and not isinstance(experiment_id, str):
            raise EndpointClientError(
                "'experiment_id' must be of type 'str', got '{0}'".format(type(experiment_id)))
        if message_type is not None and not isinstance(message_type, str):
            raise EndpointClientError(
                "'stimulus_id' must be of type 'str', got '{0}'".format(type(stimulus_id)))

        message = Message(message_type,
                          request_body.get('payload', None),
                          experiment_id,
                          stimulus_id)

        self._device_coordinator.dispatch(message)

        return "message dispatched"



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import platform
from typing import List, Optional
from brainflow.board_shim import BrainFlowInputParams
from octopus_sensing.devices.brainflow_streaming import BrainFlowStreaming

class BrainFlowOpenBCIStreaming(BrainFlowStreaming):
    '''
    Manages OpenBCI streaming using brainflow
    Data will be recorded in a csv file/files with the following column order:
    channels, Acc_x, Acc_y, Acc_z, sample_id, time_stamp, trigger

    Attributes
    -----------

    Parameters
    ----------
    name: str, default: None
        Device name. This name will be used in the output path to identify 
        each device's data.

    output_path: str,  default: output
        The path for recording files.
        Audio files will be recorded in folder {output_path}/{name}
    
    saving_mode: int, default: SavingModeEnum.CONTINIOUS_SAVING_MODE
        The way of saving data: saving continiously in a file or save data related to
        each stimulus in a separate file. 
        SavingModeEnum is:

            0. CONTINIOUS_SAVING_MODE
            1. SEPARATED_SAVING_MODE
    
    board_type: str, default: cyton-daisy
        The type of OpenBCI boards that connect by USB dongle.
        It can be:

            - cyton: for cyton board sampling rate is 250 and it has 8 channels
            - cyton-daisy: for cyton-daisy board sampling rate is 125 and it has 16 channels
            - ganglion: for Ganglion board sampling rate is 200 and it has 4 channels
    
    serial_port: str, default: None
        The serial port for reading OpenBCI data. By default we set this as follows for various platforms:

            - Linux: /dev/ttyUSB0
            - Windows: Com3
            - MacOS: /dev/cu.*

    channels_order: List(str), default: None
        A list of channel names which specify the order and names of channels

    Example
    --------
    Creating an instance of OpenBCI board with USB dongle using 
    `brainflow <https://brainflow.readthedocs.io/en/stable/SupportedBoards.html#openbci>`_,
    and adding it to the device coordinator. Device coordinator is responsible
    for triggerng the OpenBCI to start or stop recording  or to add markers to
    recorded data.
    In this example, since the saving mode is continuous, all recorded data 
    will be saved in a file. But, when an event happens, device coordinator will send
    a trigger message to the device and recorded data will be marked with the trigger
    
    >>> my_openbci =
    ...     BrainFlowOpenBCIStreaming(name="OpenBCI",
    ...                               output_path="./output",
    ...                               board_type="cyton-daisy",
    ...                               saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE,
    ...                               channels_order=["Fp1", "Fp2", "F7", "F3", 
    ...                                               "F4", "F8", "T3", "C3",
    ...                                               "C4", "T4", "T5", "P3", 
    ...                                               "P4", "T6", "O1", "O2"])
    >>> device_coordinator.add_device(my_openbci)

    Note
    -----
    Before running the code, turn on the OpenBCI, connect the dongle and make sure its port is free.
  

    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`,
    :class:`octopus_sensing.devices.brainflow_streaming`
    '''

    def __init__(self,
                 channels_order: List[str]=None,
                 board_type:str ="cyton-daisy",
                 name: Optional[str] = None,
                 output_path: str = "output",
                 serial_port=None):
        self.channels = channels_order        
        if board_type == "cyton-daisy":
            device_id = 2
            sampling_rate = 125
            if self.channels is None:
                self.channels = \
                    ["ch1", "ch2", "ch3", "ch4", "ch5", "ch6", "ch7", "ch8", "ch9",
                    "ch10", "ch11", "ch12", "ch13", "ch14", "ch15", "ch16"]
            if len(self.channels) != 16:
                raise RuntimeError("The number of channels in channels_order should be 16")
        elif board_type == "cyton":
            device_id = 0
            sampling_rate = 250
            self.channels = \
                ["ch1", "ch2", "ch3", "ch4", "ch5", "ch6", "ch7", "ch8"]
            if len(self.channels) != 8:
                raise RuntimeError("The number of channels in channels_order should be 8")
        elif board_type == "Ganglion":
            device_id = 1
            sampling_rate = 200
            self.channels = \
                ["ch1", "ch2", "ch3", "ch4"]
            if len(self.channels) != 4:
                raise RuntimeError("The number of channels in channels_order should be 4")
        else:
            raise RuntimeError("Use BrainflowStreaming fr other boards")

        if serial_port is None:
            if platform.system() == "Linux":
                serial_port = "/dev/ttyUSB0"
            elif platform.system() == "Windows":
                serial_port = "Com3"
            else:
                serial_port = "/dev/cu.*"

        params = BrainFlowInputParams()
        params.serial_port = serial_port
        super().__init__(device_id,
                         sampling_rate,
                         brain_flow_input_params=params,
                         name=name,
                         output_path=output_path)
    def get_output_path(self):
        '''
        Gets the path that is used for data recording

        Returns
        -----------
        output_path: str
           The output path that use for data recording
        '''
        return self.output_path

    def get_channels(self):
        '''
        Gets the list of channels

        Returns
        -------

        channels_name: List[str]
            The list of channels' name

        '''
        return self.channels
    
    def get_saving_mode(self):
        '''
        Gets saving mode
        
        Returns
        -----------
        saving_mode: int
            The way of saving data: saving continiously in a file or save data related to
            each stimulus in a separate file. 
            SavingModeEnum is CONTINIOUS_SAVING_MODE = 0 or SEPARATED_SAVING_MODE = 1
        '''
        return self._saving_mode


# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.


class SavingModeEnum():
    '''
    In CONTINIOUS_SAVING_MODE all data will be saved in one file, and some markers
    will be used to specify an event. In SEPARATED_SAVING_MODE, the data related to different events
    will be saved in separate files. 
    For example, if stimuli is a list of videos, in CONTINIOUS_SAVING_MODE, the data recorded
    during displaying all videos, will be recorded in a file and two markers one for start 
    and one for the stop of each video will be recorded in the data. In SEPARATED_SAVING_MODE for 
    each video stimuli, the data recorded while displaying each video will be recorded in 
    a separated file.


    '''
    CONTINIOUS_SAVING_MODE = 0
    SEPARATED_SAVING_MODE = 1



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import sys
import re
import threading
import http.client
import pickle
import json
from typing import Optional, Callable, List, Union

import msgpack

from octopus_sensing.common.message_creators import MessageType
from octopus_sensing.common.message import Message
from octopus_sensing.devices.device import Device


class SerializationTypes:
    JSON = json.dumps
    MSGPACK = msgpack.packb
    PICKLE = pickle.dumps


class HttpNetworkDevice(Device):
    '''
    This class can be used to send messages (triggers) to the other softwares
    through HTTP protocol.

    The external software need to listen for incoming HTTP connections (a.k.a
    HTTP server or HTTP endpoint). This device will POST message to the specified
    URLs.

    The messages will be in the form of a dictionary, serialized by the specified
    type. For example, if the `serialization_type` is JSON, a `START` message will
    be:

    {"type": "START", "experiment_id": "exp1", "stimulus_id": "s1"}

    Parameters
    ----------
    external_endpoints: List[str]
      A list of URLs to send the message to. It should have the scheme
      (http or https) at the beginning. For example: ["http://localhost:8080"]
      Note that for IPv6, you need to put it between braces, like this:
      `http://[2345:425:2ca1:0000:0000:567:5673:23b5]/'

    serialization_type:
      Should be one of the types defined in SerializationTypes

    name: str, default: None
      Name of this device

    timeout: int, default: 3
      Timeout for both connecting to the external endpoint and POSTing messages.
      In seconds.

    Example
    -------
    In your Octopus Sensing software:

    >>> coordinator = DeviceCoordinator()
    >>> device = HttpNetworkDevice(["http://localhost:8080/", "http://192.168.1.1/trigger"])
    >>> coordinator.add_device(device)

    Every message that is dispatched using `coordinator.dispatch` method will
    be send to `http://localhost:8080/`.
    '''

    def __init__(self,
                 external_endpoints: List[str],
                 serialization_type: Callable[...,
                                              Union[bytes, str]] = SerializationTypes.JSON,
                 name: Optional[str] = None,
                 timeout: int = 3):
        super().__init__(name=name)
        self._endpoints = external_endpoints
        self._timeout = timeout
        self._serialization_func = serialization_type

        # Regex is borrowed from O'Reilly website: https://www.oreilly.com/library/view/regular-expressions-cookbook/9781449327453/ch08s10.html
        # It validates and extracts host and port parts of the URL
        self._url_regex = re.compile((
            r"""([a-z][a-z0-9+\-.]*://)"""  # scheme
            r"""([a-z0-9\-._~%]+"""  # named host
            r"""|\[[a-f0-9:.]+\]"""  # ipv6 host
            r"""|\[v[a-f0-9][a-z0-9\-._~%!$&'()*+,;=:]+\])"""  # IPvFuture host
            r"""(:[0-9]+)?"""  # port
            r"""(/.*)?"""  # rest
        ))

        # We want to give these kind of errors early
        self._validate_endpoints()

    def _validate_endpoints(self):
        for endpoint in self._endpoints:
            match_object = self._url_regex.match(endpoint)
            if match_object is None:
                raise RuntimeError("HttpNetworkDevice({0}): invalid URL: {1}\nIt should be in the form http://host:port/".format(
                    self.name, endpoint))
            scheme, host, port, rest = match_object.groups()
            if scheme not in ('http://', 'https://'):
                raise RuntimeError("HttpNetworkDevice({0}): invalid URL: {1} It only supports 'http' & 'https' scheme, got: {2}".format(
                    self.name, endpoint, scheme))

    def _run(self):
        while True:
            message = self.message_queue.get()

            if message is None:
                continue

            self._send_message(message)

            if message.type == MessageType.TERMINATE:
                break

    def _send_message(self, message: Message):
        message_dict = {"type": message.type,
                        "experiment_id": message.experiment_id,
                        "stimulus_id": message.stimulus_id}
        serialized_message = self._serialization_func(message_dict)

        for endpoint in self._endpoints:
            threading.Thread(target=self._http_post,
                             args=(endpoint, serialized_message,)).start()

    def _http_post(self, endpoint: str, body: Union[bytes, str]):
        match_object = self._url_regex.match(endpoint)
        assert match_object is not None, 'We already validated all endpoints'
        scheme, host, port_str, rest = match_object.groups()

        port: Optional[int] = None
        if port_str:
            port = int(port_str)

        connect_to = "{0}:{1}".format(scheme, host)
        post_to = rest
        if not rest:
            post_to = "/"

        http_client = http.client.HTTPConnection(
            connect_to, port=port, timeout=self._timeout)

        http_client.request("POST", post_to, body=body)
        response = http_client.getresponse()

        if response.status != 200:
            print("HttpNetworkDevice({0}): Sending message to [{1}] failed: {2} {3}".format(
                  self.name, endpoint, response.status, response.reason),
                  file=sys.stderr)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any


class Message():
    '''
    Message class. 
    It is being used for communicating between DeviceCoordinator and various devices.
    DeviceCoordinator uses message objects to inform devices about an event

    Attributes
    ----------

    Parameters
    ----------
    type: str
        The type of message
    
    payload: Any
        the message data that can have differnt values
    
    experiment_id: str, default: None
        A unique ID for each participant and task in the study
    
    stimulus_id: str, default: None
        A unique ID for each stimulus
    

    Example
    -------
    We can create a customized message using this class, or use prepared messages in the common/message_creators.
    In this example, we created an instance of message, which its type is start. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices that start of `stimulus_00` is happened.

    >>> message = Message("START",
    ...                   "study_1_p10",
    ...                   "stimulus_00")
    >>> device_coordinator.dispatch(message)
    '''

    def __init__(self, message_type: str, payload: Any,
                 experiment_id: Optional[str] = None,
                 stimulus_id: Optional[str] = None):
        self.type = message_type
        self.payload = payload
        self.experiment_id = experiment_id
        self.stimulus_id = stimulus_id



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from octopus_sensing.devices.shimmer3_streaming import Shimmer3Streaming
from octopus_sensing.devices.brainflow_openbci_streaming import BrainFlowOpenBCIStreaming
from octopus_sensing.devices.camera_streaming import CameraStreaming
from octopus_sensing.devices.audio_streaming import AudioStreaming
from octopus_sensing.devices.open_vibe_streaming import OpenVibeStreaming
from octopus_sensing.devices.brainflow_streaming import BrainFlowStreaming
from octopus_sensing.devices.network_devices.socket_device import SocketNetworkDevice


# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from typing import Optional
import multiprocessing
from multiprocessing.queues import Queue


class Device(multiprocessing.Process):
    '''
    All devices should inherit from Device class.
    
    Attributes
    ----------

    Parameters
    ----------
    name: str, default: None
        Device name. This name will be used in the output path to identify 
        each device's data.

    output_path: str,  default: output
        The path for recording files.
        Audio files will be recorded in folder {output_path}/{name}   
    '''

    def __init__(self, name: Optional[str] = None, output_path: str = "output"):
        super().__init__(name=name)
        self.message_queue: Optional[Queue] = None
        self.subject_id: Optional[str] = None
        self.stimulus_id: Optional[str] = None
        self.output_path: str = output_path

    def run(self) -> None:
        '''
        Starts the device's process. DeviceCoordinator calls this mehod to start
        data recording
        '''
        self._run()

    def _run(self) -> None:
        '''The subclass shouldn't implement 'run', but '_run' instead.'''
        raise NotImplementedError()

    def set_queue(self, queue: Queue) -> None:
        '''
        Sets message queue for the device. 
        Device coordinator calls this method and set the queue for its device list

        Parameters
        -----------
        queue: Queue
            a queue that will be used for message passing
        '''
        self.message_queue = queue

    def get_name(self) -> str:
        '''
        Gets device's name

        Returns
        --------
        str
            The name of device
        '''
        return self.name



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.
import os

from octopus_sensing.devices.shimmer3_streaming import Shimmer3Streaming
from octopus_sensing.devices.openbci_streaming import OpenBCIStreaming
from octopus_sensing.devices.common import SavingModeEnum
from octopus_sensing.preprocessing.preprocess_devices import preprocess_devices
from octopus_sensing.device_coordinator import DeviceCoordinator

RECORDED_FILES_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   "data/recorded")
print("RECORDED_FILES_PATH", RECORDED_FILES_PATH)


class MockedShimmer3Streaming(Shimmer3Streaming):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _run(self):
        pass

    def _inintialize_connection(self):
        pass

    def _make_output_path(self):
        return self.output_path


class MockedOpenBCIStreaming(OpenBCIStreaming):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _run(self):
        pass

    def _inintialize_board(self, daisy):
        return None

    def _make_output_path(self):
        return self.output_path


def test_preprocess_continuous_saving_mode():
    ch_names = ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3",
                "C4", "T4", "T5", "P3", "P4", "T6", "O1", "O2"]
    openbci16 = \
        MockedOpenBCIStreaming(name="OpenBCI_16_continuous",
                               saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                               output_path=os.path.join(
                                   RECORDED_FILES_PATH, "OpenBCI_16_continuous"),
                               channels_order=ch_names)
    ch_names = ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3"]
    openbci8 = \
        MockedOpenBCIStreaming(name="OpenBCI_8_continuous",
                               saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                               output_path=os.path.join(
                                   RECORDED_FILES_PATH, "OpenBCI_8_continuous"),
                               channels_order=ch_names,
                               daisy=False)

    shimmer = \
        MockedShimmer3Streaming(name="Shimmer_continuous",
                                saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                                output_path=os.path.join(RECORDED_FILES_PATH, "Shimmer_continuous"))

    device_coordinator = DeviceCoordinator()
    device_coordinator.add_devices([openbci16, openbci8, shimmer])

    preprocess_file_path = "/tmp/preprocess"

    preprocess_devices(device_coordinator, preprocess_file_path,
                       openbci_sampling_rate=6,
                       shimmer3_sampling_rate=6,
                       signal_preprocess=False)

    expected_openbci16_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/OpenBCI_16_continuous")
    preprocess_openbci16_path = "/tmp/preprocess/OpenBCI_16_continuous"
    check_files(preprocess_openbci16_path, expected_openbci16_path)

    expected_openbci8_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/OpenBCI_8_continuous")
    preprocess_openbci8_path = "/tmp/preprocess/OpenBCI_8_continuous"
    check_files(preprocess_openbci8_path, expected_openbci8_path)

    expected_shimmer_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/Shimmer_continuous/gsr")
    preprocess_shimmer_path = "/tmp/preprocess/Shimmer_continuous/gsr"
    check_files(preprocess_shimmer_path, expected_shimmer_path)

    expected_shimmer_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/Shimmer_continuous/ppg")
    preprocess_shimmer_path = "/tmp/preprocess/Shimmer_continuous/ppg"
    check_files(preprocess_shimmer_path, expected_shimmer_path)


def test_preprocess_seperated_saving_mode():
    ch_names = ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3",
                "C4", "T4", "T5", "P3", "P4", "T6", "O1", "O2"]
    openbci16 = \
        MockedOpenBCIStreaming(name="OpenBCI_16_sep",
                               saving_mode=SavingModeEnum.SEPARATED_SAVING_MODE,
                               output_path=os.path.join(
                                   RECORDED_FILES_PATH, "OpenBCI_16_sep"),
                               channels_order=ch_names)
    ch_names = ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3"]
    openbci8 = \
        MockedOpenBCIStreaming(name="OpenBCI_8_sep",
                               saving_mode=SavingModeEnum.SEPARATED_SAVING_MODE,
                               output_path=os.path.join(
                                   RECORDED_FILES_PATH, "OpenBCI_8_sep"),
                               channels_order=ch_names,
                               daisy=False)
    shimmer = \
        MockedShimmer3Streaming(name="Shimmer_sep",
                                saving_mode=SavingModeEnum.SEPARATED_SAVING_MODE,
                                output_path=os.path.join(RECORDED_FILES_PATH, "Shimmer_sep"))

    device_coordinator = DeviceCoordinator()
    device_coordinator.add_devices([openbci16, openbci8, shimmer])

    preprocess_file_path = "/tmp/preprocess"
    preprocess_devices(device_coordinator, preprocess_file_path,
                       openbci_sampling_rate=6,
                       shimmer3_sampling_rate=6,
                       signal_preprocess=False)

    expected_openbci16_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/OpenBCI_16_sep")
    preprocess_openbci16_path = "/tmp/preprocess/OpenBCI_16_sep"
    check_files(preprocess_openbci16_path, expected_openbci16_path)

    expected_openbci8_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/OpenBCI_8_sep")
    preprocess_openbci8_path = "/tmp/preprocess/OpenBCI_8_sep"
    check_files(preprocess_openbci8_path, expected_openbci8_path)

    # For GSR
    expected_shimmer_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/Shimmer_sep/gsr")
    preprocess_shimmer_path = "/tmp/preprocess/Shimmer_sep/gsr"
    check_files(preprocess_shimmer_path, expected_shimmer_path)

    #For PPG
    expected_shimmer_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     "data/preprocess_expected/Shimmer_sep/ppg")
    preprocess_shimmer_path = "/tmp/preprocess/Shimmer_sep/ppg"
    check_files(preprocess_shimmer_path, expected_shimmer_path)


def check_files(preprocess_path, expected_path):
    expected_files = os.listdir(expected_path)
    preprocess_files = os.listdir(preprocess_path)
    assert len(preprocess_files) == len(expected_files)

    expected_files.sort()
    preprocess_files.sort()
    for file in preprocess_files:
        # Compare file content
        actual_file = open(os.path.join(preprocess_path, file), 'r').read()
        expected_file = open(os.path.join(expected_path, file), 'r').read()
        assert actual_file == expected_file



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os
import heartpy as hp
from scipy import signal
import numpy as np
import pathlib

from octopus_sensing.preprocessing.utils import load_all_trials, resample, load_all_samples
from octopus_sensing.devices.common import SavingModeEnum


def shimmer3_preprocess(input_path: str, file_name: str, output_path: str,
                        saving_mode: int = SavingModeEnum.CONTINIOUS_SAVING_MODE,
                        sampling_rate: int = 128,
                        signal_preprocess: bool = True):
    '''
    Preprocess shimmer recorded files to prepare them for visualizing and analysis
    It applys data cleaning (according to signal_preprocess), resampling (according to sampling_rate),
    and splits data if data has been recorded continuously. It will save PPG and GSR data in separated files

    Parameters
    ----------
    input_path: str
        The path to recorded shimmer data
    
    file_name: str
        The file name of recorded shimmer data
    
    output_path: str
        preprocessed file path
    
    saving_mode: int, default: SavingModeEnum.CONTINIOUS_SAVING_MODE
        The saving mode of recorded data. If it is CONTINIOUS_SAVING_MODE, data will be splitted
        according to markers and will be recorded in the separated files

    sampling_rate: int, default: 128
        The desired sampling_rate. Data will be resampled according to this sampling rate
    
    signal_preprocess: bool, default: True
        If True will apply preliminary preprocessing steps to clean line noises
    
    Note
    -----
    Sometimes recorded data in one second with Shimmer3 are less or more than 
    the specified sampling rate. So, we resample data by replicating
    the last samples or removing some samples to achieve the desired sampling_rate
    '''
    if saving_mode == SavingModeEnum.SEPARATED_SAVING_MODE:
        data, times = \
            load_all_samples(os.path.join(input_path, file_name),
                             (5, 7),
                             7,
                             '%Y-%m-%d %H:%M:%S.%f')
        resampled_data = \
            resample(data, times, sampling_rate)
        if signal_preprocess is True:
            cleaned_gsr = \
                clean_gsr(resampled_data[:, 0],
                          sampling_rate)
            cleaned_ppg = \
                clean_ppg(resampled_data[:, 1],
                          sampling_rate)
        else:
            cleaned_gsr = resampled_data[:, 0]
            cleaned_ppg = resampled_data[:, 1]

        gsr_output_path = os.path.join(output_path, "gsr")
        if not os.path.exists(gsr_output_path):
            pathlib.Path(gsr_output_path).mkdir(parents=True, exist_ok=True)
        gsr_file_path = \
            "{0}/gsr{1}".format(gsr_output_path, file_name[7:])
        
        ppg_output_path = os.path.join(output_path, "ppg")
        if not os.path.exists(ppg_output_path):
            pathlib.Path(ppg_output_path).mkdir(parents=True, exist_ok=True)
        ppg_file_path = \
            "{0}/ppg{1}".format(ppg_output_path, file_name[7:])
        np.savetxt(gsr_file_path, cleaned_gsr)
        np.savetxt(ppg_file_path, cleaned_ppg)

    elif saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
        print("shimmer input_path", input_path)
        # First data needs to be splitted based on markers
        trials_data, trials_times, triger_list = \
            load_all_trials(os.path.join(input_path, file_name),  # File path
                            (5, 7),  # channel columns
                            7,  # timestamp column
                            8,  # triger column
                            '%Y-%m-%d %H:%M:%S.%f')  # timestamp format

        i = 0
        for trial in trials_data:
            gsr_output_path = os.path.join(output_path, "gsr")
            if not os.path.exists(gsr_output_path):
                pathlib.Path(gsr_output_path).mkdir(parents=True, exist_ok=True)
            gsr_file_path = \
                "{0}/gsr{1}-{2}.csv".format(gsr_output_path,
                                            # Removing .csv and shimmer from file_name
                                            file_name[7:-4],
                                            str(triger_list[i]).zfill(2))
            ppg_output_path = os.path.join(output_path, "ppg")
            if not os.path.exists(ppg_output_path):
                pathlib.Path(ppg_output_path).mkdir(parents=True, exist_ok=True)
            ppg_file_path = \
                "{0}/ppg{1}-{2}.csv".format(ppg_output_path,
                                            # Removing .csv and shimmer from file_name
                                            file_name[7:-4],
                                            str(triger_list[i]).zfill(2))

            resampled_data = \
                resample(trial, trials_times[i], sampling_rate)

            if signal_preprocess is True:
                print("shape", resampled_data.shape)
                cleaned_gsr = \
                    clean_gsr(resampled_data[:, 0],
                              sampling_rate)
                cleaned_ppg = \
                    clean_ppg(resampled_data[:, 1],
                              sampling_rate)
            else:
                cleaned_gsr = resampled_data[:, 0]
                cleaned_ppg = resampled_data[:, 1]

            np.savetxt(gsr_file_path, cleaned_gsr)
            np.savetxt(ppg_file_path, cleaned_ppg)
            i += 1
    else:
        raise Exception("Saving mode is incorrect")


def clean_gsr(data, sampling_rate: int, low_pass: float=0.1, high_pass: float=15):
    '''
    Removes high frequency and rapid transient noises

    Parameters
    -----------
    data: numpy.array
        An 1D array of GSR data

    smpling_rate: int, default: 128
        sampling rate

    low_pass: float, default: 0.7
        The low cut frequency for filtering
    
    high_pass: float, default: 2.5
        The high cut frequency for filtering
    
    Returns
    -------
    cleaned_data: numpy.array
        An 1D array of cleaned GSR data
    '''
    nyqs = sampling_rate * 0.5
    # Removing high frequency noises
    b, a = signal.butter(5, [low_pass / nyqs, high_pass / nyqs], 'bands')
    output = signal.filtfilt(b, a, np.array(data, dtype=np.float64))

    # Removing rapid transient artifacts
    final_output = signal.medfilt(output, kernel_size=5)
    return final_output


def clean_ppg(data: np.ndarray, sampling_rate: int, low_pass: float=0.7, high_pass: float=2.5):
    '''
    Removes high frequency noises

    It uses `heartpy <https://github.com/paulvangentcom/heartrate_analysis_python>` library

    Parameters
    -----------
    data: numpy.ndarray
        An 1D array of PPG data

    smpling_rate: int, default: 128
        sampling rate

    low_pass: float, default: 0.7
        The low cut frequency for filtering
    
    high_pass: float, default: 2.5
        The high cut frequency for filtering
    
    Returns
    -------
    cleaned_data: numpy.array
        An 1D array of cleaned PPG data

    '''
    filtered = hp.filter_signal(data,
                                [low_pass, high_pass],
                                sample_rate=sampling_rate,
                                order=3,
                                filtertype='bandpass')

    return filtered



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import os
import pathlib
from typing import List, Dict
from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.devices.openbci_streaming import OpenBCIStreaming
from octopus_sensing.devices import BrainFlowOpenBCIStreaming
from octopus_sensing.devices import Shimmer3Streaming
from octopus_sensing.preprocessing.openbci import openbci_preprocess
from octopus_sensing.preprocessing.openbci_brainflow import openbci_brainflow_preprocess
from octopus_sensing.preprocessing.shimmer3 import shimmer3_preprocess


def preprocess_devices(device_coordinator: DeviceCoordinator, output_path: str,
                       openbci_sampling_rate: int = 128,
                       shimmer3_sampling_rate: int = 128,
                       signal_preprocess: bool = True):
    '''
    Preprocees recorded files for all devices that are added to device_coordinator and has a 
    preprocessing module. Some devices do not have any preprocessing, so this function will ignore them

    Parameters
    ----------
    device_coordinator: DeviceCoordinator
        an instance of DeviceCoordinator
    
    output_path: str
        Path for preprocessed Files
    
    openbci_sampling_rate: int
        New sampling rate for openbci resampling
    
    shimmer3_sampling_rate: int
        New sampling rate for shimmer3 resampling
    '''
    print("Start preprocessing ....")
    devices = device_coordinator.get_devices()
    for device in devices:
        if isinstance(device, OpenBCIStreaming):
            device_output_path = os.path.join(output_path, device.get_name())
            print("device_output_path", device_output_path)
            if not os.path.exists(device_output_path):
                pathlib.Path(device_output_path).mkdir(parents=True, exist_ok=True)

            # This is the path that device saves recording data
            input_path = device.get_output_path()
            print("input_path", input_path)
            file_names = os.listdir(input_path)
            file_names.sort()
            if not os.path.exists(device_output_path):
                os.mkdir(device_output_path)
            for file_name in file_names:
                openbci_preprocess(input_path, file_name, device_output_path,
                                   device.get_channels(),
                                   saving_mode=device.get_saving_mode(),
                                   sampling_rate=openbci_sampling_rate,
                                   signal_preprocess=signal_preprocess)
        elif isinstance(device, BrainFlowOpenBCIStreaming):
            device_output_path = os.path.join(output_path, device.get_name())
            print("device_output_path", device_output_path)
            if not os.path.exists(device_output_path):
                pathlib.Path(device_output_path).mkdir(parents=True, exist_ok=True)

            # This is the path that device saves recording data
            input_path = device.get_output_path()
            print("input_path", input_path)
            file_names = os.listdir(input_path)
            file_names.sort()
            if not os.path.exists(device_output_path):
                os.mkdir(device_output_path)
            for file_name in file_names:
                openbci_brainflow_preprocess(input_path, file_name, device_output_path,
                                             device.get_channels(),
                                             saving_mode=device.get_saving_mode(),
                                             sampling_rate=openbci_sampling_rate,
                                             signal_preprocess=signal_preprocess)
        elif isinstance(device, Shimmer3Streaming):
            device_output_path = os.path.join(output_path, device.get_name())
            if not os.path.exists(device_output_path):
                pathlib.Path(device_output_path).mkdir(parents=True, exist_ok=True)

            # This is the path that device saves recording data

            input_path = device.get_output_path()
            file_names = os.listdir(input_path)
            file_names.sort()
            print("preprocess shimmer input_path", input_path)
            print("preprocess shimmer device_output_path", device_output_path)
            for file_name in file_names:
                shimmer3_preprocess(input_path, file_name, device_output_path,
                                    saving_mode=device.get_saving_mode(),
                                    sampling_rate=shimmer3_sampling_rate,
                                    signal_preprocess=signal_preprocess)
    print("Preprocessing done")


def preprocess_devices_by_path(devices_path: Dict[str, str], output_path: str,
                       openbci_channels: List[str] = ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3",
                        "C4", "T4", "T5", "P3", "P4", "T6", "O1", "O2"],
                       openbci_sampling_rate: int = 128,
                       shimmer3_sampling_rate: int = 128,
                       signal_preprocess: bool = True):
    '''
    Gets a list of path to the recorded data from different devices and preprocess them if they have  
    preprocessing module. Some devices do not have any preprocessing, so this function will ignore them

    Parameters
    ----------
    devices_path: dict[str, str]
        a dictionary of device_name: data_path that specify the path to recorded data by each device
       
    output_path: str
        Path for preprocessed Files
    
    openbci_channels: List[str]
        default is ["Fp1", "Fp2", "F7", "F3", "F4", "F8", "T3", "C3", "C4", "T4", "T5", "P3", "P4", "T6", "O1", "O2"]
        A list of OpenBCI channels' names
    
    openbci_sampling_rate: int
        New sampling rate for openbci resampling
    
    shimmer3_sampling_rate: int
        New sampling rate for shimmer3 resampling
    '''

    print("Start preprocessing ....")
    for device, input_path in devices_path.items():
        print(device, input_path)
        if device == "openbci":
            device_output_path = os.path.join(output_path, "openbci")
            print("device_output_path", device_output_path)
            if not os.path.exists(device_output_path):
                pathlib.Path(device_output_path).mkdir(parents=True, exist_ok=True)

            # This is the path that device saves recording data
            print("input_path", input_path)
            file_names = os.listdir(input_path)
            file_names.sort()
            if not os.path.exists(device_output_path):
                os.mkdir(device_output_path)
            for file_name in file_names:
                print(file_name, device_output_path)
                openbci_preprocess(input_path, file_name, device_output_path,
                                   openbci_channels,
                                   sampling_rate=openbci_sampling_rate,
                                   signal_preprocess=signal_preprocess)
        elif device == "openbci_brainflow":
            device_output_path = os.path.join(output_path, "openbci_brainflow")
            print("device_output_path", device_output_path)
            if not os.path.exists(device_output_path):
                pathlib.Path(device_output_path).mkdir(parents=True, exist_ok=True)

            # This is the path that device saves recording data
            print("input_path", input_path)
            file_names = os.listdir(input_path)
            file_names.sort()
            if not os.path.exists(device_output_path):
                os.mkdir(device_output_path)
            for file_name in file_names:
                print(file_name, device_output_path)
                openbci_brainflow_preprocess(input_path, file_name, device_output_path,
                                             openbci_channels,
                                             sampling_rate=openbci_sampling_rate,
                                             signal_preprocess=signal_preprocess)
        elif device == "shimmer3":
            device_output_path = output_path
            print("device_output_path", device_output_path)
            if not os.path.exists(device_output_path):
                pathlib.Path(device_output_path).mkdir(parents=True, exist_ok=True)

            # This is the path that device saves recording data
            print("input_path", input_path)
            if not os.path.exists(device_output_path):
                pathlib.Path(device_output_path).mkdir(parents=True, exist_ok=True)
            file_names = os.listdir(input_path)
            file_names.sort()
            print("preprocess shimmer input_path", input_path)
            print("preprocess shimmer device_output_path", device_output_path)
            for file_name in file_names:
                shimmer3_preprocess(input_path, file_name, device_output_path,
                                    sampling_rate=shimmer3_sampling_rate,
                                    signal_preprocess=signal_preprocess)
    print("Preprocessing done")



import pytest
import multiprocessing
import multiprocessing.queues
import threading
import queue
import http.client
import pickle
import os
import json
import numpy as np
import time
import tempfile
from brainflow import board_shim

import octopus_sensing.devices.brainflow_streaming as brainflow_streaming
from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.common.message_creators import start_message, stop_message, terminate_message
from octopus_sensing.monitoring_endpoint import MonitoringEndpoint


class MockBrainFlowInputParams():
    def __init__(self):
        pass

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

class MockBoardShim():
    def __init__(self, board_id, input_params):
        try:
            self.input_json = input_params.to_json().encode()
        except BaseException:
            self.input_json = input_params.to_json()
        self.board_id = board_id
        self._master_board_id = self.board_id

    def set_log_level(self, a):
        pass
    def prepare_session(self):
        pass
    def start_stream(self):
        pass
    def stop_stream(self):
        pass

    def get_board_data(self):
        return (np.array([ [1,1],[1,1],[1,1] ]))


@pytest.fixture(scope="module")
def mocked():
    original_bases = brainflow_streaming.BrainFlowStreaming.__bases__[0].__bases__[0].__bases__
    brainflow_streaming.BrainFlowStreaming.__bases__[0].__bases__[0].__bases__ = (threading.Thread,)
    original_BrainFlowInputParams = brainflow_streaming.BrainFlowInputParams
    brainflow_streaming.BrainFlowInputParams = MockBrainFlowInputParams
    original_BoardShim = brainflow_streaming.BoardShim
    brainflow_streaming.BoardShim = MockBoardShim

    yield None

    brainflow_streaming.BrainFlowStreaming.__bases__[0].__bases__[0].__bases__ = original_bases
    brainflow_streaming.BrainFlowInputParams = original_BrainFlowInputParams
    brainflow_streaming.BoardShim = original_BoardShim

def test_system_health(mocked):

    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")
    experiment_id = 'test-exp-2'
    stimuli_id = 'sti-2'

    params = board_shim.BrainFlowInputParams()
    params.serial_port = "/dev/ttyUSB0"
    device = \
        brainflow_streaming.BrainFlowStreaming(2,
                                               125,
                                               brain_flow_input_params=params,
                                               name="cyton_daisy",
                                               output_path=output_dir)    
    msg_queue = queue.Queue()
    device.set_queue(msg_queue)
    monitoring_queue_in = queue.Queue()
    monitoring_queue_out = queue.Queue()
    device.set_monitoring_queues(monitoring_queue_in, monitoring_queue_out)

    device.start()

    time.sleep(0.2)

    msg_queue.put(start_message(experiment_id, stimuli_id))
    # Allowing data collection for one second
    time.sleep(1)

    msg_queue.put(stop_message(experiment_id, stimuli_id))

    time.sleep(0.2)

    # Sending terminate and waiting for the device process to exit.
    msg_queue.put(terminate_message())
    device.join()

    # It should save the file after receiving a TERMINATE.
    brain_output = os.path.join(output_dir, "cyton_daisy")
    filename = "cyton_daisy-{}.csv".format(experiment_id)

    assert os.path.exists(brain_output)
    assert len(os.listdir(brain_output)) == 1
    assert os.listdir(brain_output)[0] == filename

    filecontent = open(os.path.join(brain_output, filename), 'r').read()
    assert len(filecontent) >= 375
    # TODO: Check if the triggers are there.
    # TODO: We can check data in monitoring queues as well.
    


# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Any, Optional, Union
import os
import threading
import cv2
from octopus_sensing.devices.device import Device
from octopus_sensing.common.message_creators import MessageType
import time

class CameraStreaming(Device):
    '''
    Stream and Record video data. 
    If we have several stimuli, one vide file will be recorded for each stimuli.
    Device coordinator is responsible for triggerng the camera to start or stop recording.
    The content of recorded file is the recorded video between start and stop triggers

    Attributes
    ----------

    Parameters
    ----------
    name: str, default: None:
        The name of device
    
    output_path: str, default: output
                The path for recording files.
                Audio files will be recorded in folder {output_path}/{name}

    camera_no: int, default:0
        The camera number. Default is 0  which is defaul camera in system
    
    camera_path: str, default: None
        The physical path of camera device. It varies in different platforms.
        For Example in linux it can be something like this:
        `/dev/v4l/by-id/usb-046d_081b_97E6A7D0-video-index0`
    
    image_width: int, default: 1280
        The width of recorded frame/frames

    image_height: int, default: 720
        The height of recorded frame/frames.
    

    Notes
    -----
    - Only one of camera_no or camera_path should have value.

    - There is no guarantee that we can set the camera resolution. 
      Because camera may not be able to support these resolution and it will change it
      based on its settings


    Example
    -----------
    Creating an instance of camera and adding it to the device coordinator.
    Device coordinator is responsible for triggerng the camera to start or stop recording

    >>> camera = CameraStreaming(camera_no=0,
    ...                          name="camera",
    ...                          output_path="./output")
    >>> device_coordinator.add_device(camera)

    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`

    '''
    def __init__(self, camera_no: Optional[Union[int, str]] = None,
                 camera_path: Optional[str] = None,
                 image_width: int = 1280,
                 image_height: int = 720,
                 **kwargs):
        assert (camera_no is not None) ^ (camera_path is not None), \
            "Only one of camera_no or camera_path should have value"
        super().__init__(**kwargs)
        self.output_path = os.path.join(self.output_path, self.name)
        os.makedirs(self.output_path, exist_ok=True)
        if camera_no is not None:
            self._camera_number = camera_no
        elif camera_path is not None:
            self._camera_number = os.path.realpath(camera_path)

        self._image_width = image_width
        self._image_height = image_height
        self._video_size: Tuple[int, int] = (self._image_width, self._image_height)
        self._video_capture: Any = None
        self._fps: int = 30
        self._capture_times: list = []
        self._frames: list = []
        self._counter = 0
        self._state = ""

    def _run(self):
        print("self._camera_number", self._camera_number)
        self._video_capture = cv2.VideoCapture(self._camera_number)
        try:
            self._video_capture.set(cv2.CAP_PROP_FRAME_WIDTH, self._image_width)
            self._video_capture.set(cv2.CAP_PROP_FRAME_HEIGHT, self._image_height)
        except Exception as error:
            # Ignoring all errors
            print("Could not set the camera resolution.")
            print(error)

        # There's no guarantee that we can set the camera resolution. So, we
        # re-read the settings again from the camera.
        self._video_size = (int(self._video_capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
                            int(self._video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        self._video_capture.read()

        recording_thread = None
        recording_event = None

        print(f"Initialized video device [{self.name}]: {self._video_size}")

        while True:
            message = self.message_queue.get()
            if message is None:
                continue
            if message.type == MessageType.START:
                if self._state == "START":
                    print("Video streaming has already started")
                else:
                    self._frames = []
                    self._capture_times = []
                    if recording_thread is not None:
                        raise RuntimeError(
                            ("[{0} device] Received two start messages. "
                            "A STOP message should be send before trying "
                            "to start a new video recording.".format(self.name)))

                    file_name = "{0}/{1}-{2}-{3}.avi".format(self.output_path,
                                                            self.name,
                                                            message.experiment_id,
                                                            str(message.stimulus_id).zfill(2))
                    recording_event = threading.Event()
                    recording_event.set()
                    recording_thread = threading.Thread(
                        target=self._stream_loop, args=(file_name, recording_event), daemon=True)
                    recording_thread.start()
                    self._state = "START"

            elif message.type == MessageType.STOP:
                if self._state == "STOP":
                    print("Video streaming has already started")
                else:
                    if recording_event is not None:
                        recording_event.clear()
                    recording_thread = None
                    recording_event = None
                    self._state = "STOP"

            elif message.type == MessageType.TERMINATE:
                if recording_event is not None:
                    recording_event.clear()
                recording_thread = None
                recording_event = None
                break

        print("video terminated")
        self._video_capture.release()

    def _stream_loop(self, file_name: str, event: threading.Event):
        codec = cv2.VideoWriter_fourcc(*'XVID')
        try:
            while self._video_capture.isOpened:
                if event.is_set():
                    ret, frame = self._video_capture.read()
                    if ret:
                        self._counter += 1
                        self._capture_times.append(time.time())
                        self._frames.append(frame)
                else:
                    time_diff = self._capture_times[-1]-self._capture_times[0]
                    fps = \
                        int(len(self._frames)/(time_diff))
                    print("Recording frame per second", fps)
                    writer = cv2.VideoWriter(file_name,
                            codec,
                            fps,
                            self._video_size)
                    for frame in self._frames:
                        writer.write(frame)
                    writer.release()
                    break

        except Exception as error:
            print("Error while recording video. Device: {0}".format(self.name))
            print(error)

    def _stream_loop_image(self, file_name: str, event: threading.Event):
        try:
            while self._video_capture.isOpened:
                if event.is_set():
                    ret, frame = self._video_capture.read()
                    if ret:
                        a = time.time()
                        os.makedirs(file_name[:-4], exist_ok=True)
                        image_file_name = file_name[:-4] + "/" + str(a) + ".jpg"
                        print(image_file_name)
                        cv2.imwrite(image_file_name, frame)
                else:
                    break

        except Exception as error:
            print("Error while recording video. Device: {0}".format(self.name))
            print(error)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import threading
from gi.repository import Gtk, Gst
import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Gst', '1.0')


from octopus_sensing.windows import ImageWindow
from octopus_sensing.stimuli.stimulus import Stimulus

Gst.init(None)
Gst.init_check(None)

class ImageStimulus(Stimulus):
    '''
    Attributes
    -----------


    Parameters
    ----------
    id: str
        The id of stimulus

    path: str
        The path of image
    
    show_time: int
        The time period for displaying the image in seconds
    
    monitor_no: int, default: 0
        The ID of monitor for displaying of image. It can be 0, 1, ...
    '''

    def __init__(self, id: str, path: str, show_time: int, monitor_no:int =0):
        super().__init__(id, path)
        self._show_time = show_time
        self.image = ImageWindow(path, show_time, monitor_no=monitor_no)

    def show(self):
        '''
        If we have a main Gtk window and want to display image in it, we are using this
        method to display the image. Otherwise we use show_standalone method.
        '''
        self.image.show_window()

    def show_standalone(self):
        '''
        If we don't have a Gtk window and want to display it standalone, we are using this method. 
        It may have some miliseconds delay to initialize GTK.
        If these miliseconds are important, don't use this method to display image

        '''
        def gtk_main():
            self.image.show_window()
            self.image.connect("destroy", Gtk.main_quit)
            Gtk.main()

        threading.Thread(target=gtk_main).start()



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import sys
import time
import traceback
import multiprocessing
import queue
import pickle
from typing import List, Any, Tuple, Dict

from octopus_sensing.devices.device import Device
from octopus_sensing.devices.monitored_device import MonitoredDevice
from octopus_sensing.common.message import Message
from octopus_sensing.common.message_creators import terminate_message

QueueType = multiprocessing.queues.Queue


class MonitoringCache:
    def __init__(self):
        self._time = time.time_ns()
        self._cached_data = []

    def get_cache(self):
        '''Returns None if cache is not available or expired'''
        # 100 ms
        if time.time_ns() - self._time > 100000000:
            return None
        return self._cached_data

    def cache(self, data: List[Any]):
        self._cached_data = data
        self._time = time.time_ns()


class DeviceCoordinator:
    '''
    Device coordinator is responsible for coordination, like start recording data
    from all devices at once, stop recording, triggering (marking data at point), 
    and terminating devices. 
    When a device is added to the device coordinator, it will be initialized and prepared for data recording.
    '''

    def __init__(self):
        self.__devices = {}
        self.__queues: List[QueueType] = []
        self.__device_counter: int = 0
        self.__monitoring_queues: List[Tuple[QueueType, QueueType, Device]] = [
        ]
        self.__monitoring_cache = MonitoringCache()

    def __get_device_id(self) -> str:
        '''
        Generate an ID for devices that do not have name

        Returns
        -----------
        str
            an ID for device
        '''

        self.__device_counter += 1
        device_id = "device_{0}".format(self.__device_counter)
        return device_id

    def get_devices(self) -> List[Device]:
        '''
        Return a list of added devices

        Returns
        -----------
        List[Device]
            List of device objects
        '''

        return list(self.__devices.values())

    def add_device(self, device: Device) -> None:
        '''
        Adds new device to the coordinator and starts it

        Parameters
        ----------
        device: Device
            a device object
        
        Example
        --------
        >>> my_shimmer = Shimmer3Streaming(name="Shimmer3_sensor", output_path="./output")
        >>> device_coordinator.add_device(my_shimmer)

        See Also
        ---------
        add_devices: Adds a list of new devices to the coordinator and starts them
        '''
        assert isinstance(device, Device)

        if device.name is None:
            device.name = self.__get_device_id()
        if device.name in self.__devices.keys():
            raise RuntimeError("This device already has been added")

        self.__devices[device.name] = device
        msg_queue: QueueType = multiprocessing.Queue()
        device.set_queue(msg_queue)
        self.__queues.append(msg_queue)

        self.__set_monitoring_queues(device)

        device.start()

    def add_devices(self, devices: List[Device]) -> None:
        '''
        Adds a list of new devices to the coordinator and starts them

        Parameters
        ----------
        devices: List[Device]
            a list of device objects
    
        Example
        --------
        >>> my_shimmer = Shimmer3Streaming(name="Shimmer3_sensor",
                                           output_path="./output")
        >>> device_coordinator.add_devices([my_shimmer])

        See Also
        ---------
        add_device: Adds new device to the coordinator and starts it
        '''
        for device in devices:
            self.add_device(device)

    def dispatch(self, message: Message) -> None:
        '''
        dispatches a new message to all devices

        Parameters
        ----------
        message: Message
            a message object
        
        Example
        --------     
        >>> device_coordinator.dispatch(start_message(experiment_id,
                                                      stimuli_id))
        '''

        for message_queue in self.__queues:
            message_queue.put(message)

    def terminate(self):
        '''
        sends terminate message to all devices and terminate all processes 

        Example
        --------     
        >>> device_coordinator.terminate()
        '''
        self.dispatch(terminate_message())
        for item in self.__devices.values():
            item.join()

    def get_monitoring_data(self) -> Dict[str, List[Any]]:
        '''
        Returns latest collected data from all devices.
        Device's data can be anything, depending on the device itself.

        Returns
        ---------
        data : dict[str, list[any]]
            The keys are device names and values are collected data from the device
        
        Note
        ----
        This method is being used for real-time monitoring

        '''
        cached = self.__monitoring_cache.get_cache()
        if cached:
            return cached

        # Putting request for all devices, then collecting them all, for performance reasons.
        for in_q, _, device in self.__monitoring_queues:
            try:
                # The sub-process won't use the data we put in the queue. It's just a signal.
                in_q.put(b'0', timeout=0.1)
            except queue.Full:
                print("Could not put monitoring request for {0} device.".format(
                    device.name), file=sys.stderr)
                traceback.print_exc()

        result = {}
        for _, out_q, device in self.__monitoring_queues:
            try:
                records = pickle.loads(out_q.get(timeout=0.1))
                # We ensured device has a name in the add_device, ignoring it here.
                result[device.name] = records  # type: ignore
            except (queue.Empty, pickle.PickleError):
                print("Could not read monitoring data from {0} device".format(
                    device.name), file=sys.stderr)
                traceback.print_exc()

        self.__monitoring_cache.cache(result)
        return result

    def __set_monitoring_queues(self, device: Device) -> None:
        if isinstance(device, MonitoredDevice):
            in_q: QueueType = multiprocessing.Queue()
            out_q: QueueType = multiprocessing.Queue()
            device.set_monitoring_queues(in_q, out_q)
            self.__monitoring_queues.append((in_q, out_q, device))



import os
import time
import tempfile
import queue
import threading
import pytest

import octopus_sensing.devices.camera_streaming as camera_streaming
from octopus_sensing.common.message_creators import start_message, stop_message, terminate_message

class MockedCv2Module:

    def VideoCapture(self, camera_number):
        return MockedVideoCaptureModule(camera_number)
    
    def VideoWriter_fourcc(self, a, b, c, d):
        # choose .avi file for example
        return ('XVID')
    
    def VideoWriter(self, file_name, codec, fps, _video_size):
        return MockedVideoWriterModule(file_name)

    def CAP_PROP_FRAME_WIDTH(self):
        return (float(640.0))
    
    def CAP_PROP_FRAME_HEIGHT(self):
        return (float(480.0))
    
    def imwrite(self, filename, img):
        a = time.time()
        open(filename[:-4] + "/" + str(a) + ".jpg", 'w+').write(img)

class MockedVideoCaptureModule:

    def __init__(self, camera_number):
        pass

    def isOpened(self):
        return True

    def set(self, a, b):
        pass

    def read(self):
        # bool, 3d matrix (fake signal)
        return True, [[[0,10,10],[1,11,11]], [[0,10,10],[1,11,11]]]
    
    def release(self):
        pass

    def get(self, option):
        return (float(640.0))


class MockedVideoWriterModule:

    def __init__(self, file_name):
        self.file_name = file_name

    def write(self, frame):
        a = time.time()
        open(self.file_name, 'a').write(str(frame))

    def release(self):
        pass

@pytest.fixture(scope="module")
def mocked():
    original_bases = camera_streaming.CameraStreaming.__bases__[0].__bases__
    camera_streaming.CameraStreaming.__bases__[0].__bases__ = (threading.Thread,)

    original_cv2 = camera_streaming.cv2
    camera_streaming.cv2 = MockedCv2Module()

    yield None

    # Replacing back the original things
    camera_streaming.CameraStreaming.__bases__[0].__bases__ = original_bases
    camera_streaming.cv2 = original_cv2


def test_happy_path(mocked):

    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")
    device_name = 'test-video-device'
    experiment_id = 'test-exp-1'
    stimuli_id = 'sti-1'

    device = camera_streaming.CameraStreaming(
        camera_no=0, output_path=output_dir, name=device_name)

    msg_queue = queue.Queue()
    device.set_queue(msg_queue)
    device.start()

    time.sleep(0.2)

    msg_queue.put(start_message(experiment_id, stimuli_id))

    time.sleep(1)

    msg_queue.put(stop_message(experiment_id, stimuli_id))

    time.sleep(0.2)

    # It should save the file after receiving a STOP.
    device_output = os.path.join(output_dir, device_name)
    assert os.path.exists(device_output)

    recorded_file = os.path.join(device_output, '{}-{}-{}.avi'.format(
        device_name, experiment_id, stimuli_id))

    assert os.path.exists(recorded_file)
    assert os.path.getsize(recorded_file) > 25600 # check the file size to assure it is not empty

    # Sending terminate and waiting for the device process to exit.
    msg_queue.put(terminate_message())
    device.join()


# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import socket
import time
from octopus_sensing.devices.device import Device


class OpenVibeStreaming(Device):
    '''
    Sending triggers to OpenVibe data recorders. OpenVibe supports data acquisition through
    many biosensors. We can record data through OpenVibe and send markers using this class.

    Attributes
    ----------

    Parameters
    ----------
    host: str
        host IP address

    port: int
        port number

    Example
    -------
    Creating an instance of OpenVibeStreaming in the local machine and adding it to the device_coordinator

    >>> device_coordinator = DeviceCoordinator()
    >>> openvibe_device = OpenVibeStreaming()
    >>> device_coordinator.add_devices([openvibe_device])

    Note
    -----
    We need a scenario in OpenVibe to record data. OpenVibe should start data recording before sending the triggers.
    '''
    def __init__(self,
                 host: str='127.0.0.1',
                 port: int=15361):
        super().__init__()
        # connect
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((host, port))

    def run(self):
        while True:
            message = self.message_queue.get()
            if message is not None:
                self.subject_id = message.subject_id
                self.stimulus_id = message.stimulus_id
            if message is None:
                continue
            elif message.type == "terminate":
                break
            else:
                print("Send trigger")
                self._send_trigger(message.payload["trigger"])
        self._socket.close()

    def _send_trigger(self, event_id):
        # create the three pieces of the tag, padding, event_id and timestamp
        padding=[0]*8

        # transform the value into an array of byte values in little-endian order
        event_id = list(event_id.to_bytes(8, byteorder='little'))

        # timestamp can be either the posix time in ms, or 0 to let the acquisition server timestamp the tag itself.
        t = int(time.time()*1000)
        timestamp = list(t.to_bytes(8, byteorder='little'))

        self._socket.sendall(bytearray(padding+event_id+timestamp))
